import functools
import sys

from six import iterkeys, itervalues
from six.moves import range

from . import registers
from .. import profiler

from ..interfaces import IMachineWorker, ISnapshotable
from ..mm import ADDR_FMT, UINT8_FMT, UINT16_FMT, UINT32_FMT, SEGMENT_SIZE, PAGE_SIZE, PAGE_MASK, PAGE_SHIFT, i16, addr_to_page
from .registers import Registers, REGISTER_NAMES, FlagsRegister
from .instructions import DuckyInstructionSet
from ..errors import AccessViolationError, InvalidResourceError
from ..util import LRUCache
from ..snapshot import SnapshotNode

#: Default IVT address
DEFAULT_IVT_ADDRESS = 0x000000

#: Default size of core instruction cache, in instructions.
DEFAULT_CORE_INST_CACHE_SIZE = 256

#: Default size of core data cache, in bytes.
DEFAULT_CORE_DATA_CACHE_SIZE = 8192

#: Default data cache line length, in bytes.
DEFAULT_CORE_DATA_CACHE_LINE_LENGTH = 32

#: Default data cache associativity
DEFAULT_CORE_DATA_CACHE_LINE_ASSOC  = 4

class CPUState(SnapshotNode):
  pass

class CPUCoreState(SnapshotNode):
  def __init__(self):
    super(CPUCoreState, self).__init__('cpuid', 'coreid', 'registers', 'exit_code', 'alive', 'running', 'idle', 'ivt_address')

class InterruptVector(object):
  """
  Interrupt vector table entry.
  """

  SIZE = 4

  def __init__(self, cs = 0x00, ds = 0x00, ip = 0x0000):
    self.cs = cs
    self.ds = ds
    self.ip = ip

class CPUException(Exception):
  """
  Base class for CPU-related exceptions.

  :param string msg: message describing exceptional state.
  :param ducky.cpu.CPUCore core: CPU core that raised exception, if any.
  :param u16 ip: address of an instruction that caused exception, if any.
  """

  def __init__(self, msg, core = None, ip = None):
    super(CPUException, self).__init__(msg)

    self.core = core
    self.ip = ip

class InvalidOpcodeError(CPUException):
  """
  Raised when unknown or invalid opcode is found in instruction.

  :param int opcode: wrong opcode.
  """

  def __init__(self, opcode, *args, **kwargs):
    super(InvalidOpcodeError, self).__init__('Invalid opcode: opcode={}'.format(opcode), *args, **kwargs)

    self.opcode = opcode

class InvalidInstructionSetError(CPUException):
  """
  Raised when switch to unknown or invalid instruction set is requested.

  :param int inst_set: instruction set id.
  """

  def __init__(self, inst_set, *args, **kwargs):
    super(InvalidInstructionSetError, self).__init__('Invalid instruction set requested: inst_set={}'.format(inst_set), *args, **kwargs)

    self.inst_set = inst_set

def do_log_cpu_core_state(core, logger = None, disassemble = True):
  """
  Log state of a CPU core. Content of its registers, and other interesting or
  useful internal variables are logged.

  :param ducky.cpu.CPUCore core: core whose state should be logged.
  :param logger: called for each line of output to actualy log it. By default,
    core's :py:meth:`ducky.cpu.CPUCore.DEBUG` method is used.
  """

  logger = logger or core.DEBUG

  for i in range(0, Registers.REGISTER_SPECIAL, 4):
    regs = [(i + j) for j in range(0, 4) if (i + j) < Registers.REGISTER_SPECIAL]
    s = ['reg{:02d}={}'.format(reg, UINT16_FMT(core.registers.map[reg].value)) for reg in regs]
    logger(' '.join(s))

  logger('cs=%s    ds=%s', UINT16_FMT(core.registers.cs.value), UINT16_FMT(core.registers.ds.value))
  logger('fp=%s    sp=%s    ip=%s', UINT16_FMT(core.registers.fp.value), UINT16_FMT(core.registers.sp.value), UINT16_FMT(core.registers.ip.value))
  logger('flags=%s', core.registers.flags.to_string())
  logger('cnt=%s, alive=%s, running=%s, idle=%s, exit=%i', core.registers.cnt.value, core.alive, core.running, core.idle, core.exit_code)

  if hasattr(core, 'math_coprocessor'):
    for index, v in enumerate(core.math_coprocessor.registers.stack):
      logger('MS: %02i: %s', index, UINT32_FMT(v.value))

  if disassemble is True:
    if core.current_instruction:
      inst = core.instruction_set.disassemble_instruction(core.current_instruction)
      logger('current=%s', inst)
    else:
      logger('current=<none>')
  else:
    logger('current=<unknown>')

  for index, (ip, symbol, offset) in enumerate(core.backtrace()):
    logger('Frame #%i: %s + %s (%s)', index, symbol, offset, ip)

def log_cpu_core_state(*args, **kwargs):
  """
  This is a wrapper for ducky.cpu.do_log_cpu_core_state function. Its main
  purpose is to be removed when debug mode is not set, therefore all debug
  calls of ducky.cpu.do_log_cpu_core_state will disappear from code,
  making such code effectively "quiet".
  """

  do_log_cpu_core_state(*args, **kwargs)

class StackFrame(object):
  def __init__(self, cs, ds, fp):
    super(StackFrame, self).__init__()

    self.CS = cs
    self.DS = ds
    self.FP = fp
    self.IP = None

  def __getattribute__(self, name):
    if name == 'address':
      return self.DS * SEGMENT_SIZE * PAGE_SIZE + self.FP

    return super(StackFrame, self).__getattribute__(name)

  def __repr__(self):
    return '<StackFrame: CS={} DS={} FP={} IP={}, ({})'.format(UINT8_FMT(self.CS), UINT8_FMT(self.DS), UINT16_FMT(self.FP), UINT16_FMT(self.IP if self.IP is not None else 0), ADDR_FMT(self.address))

class InstructionCache(LRUCache):
  """
  Simple instruction cache class, based on LRU dictionary, with a limited size.

  :param ducky.cpu.CPUCore core: CPU core that owns this cache.
  :param int size: maximal number of entries this cache can store.
  """

  def __init__(self, core, size, *args, **kwargs):
    super(InstructionCache, self).__init__(core.cpu.machine.LOGGER, size, *args, **kwargs)

    self.core = core

  def get_object(self, addr):
    """
    Read instruction from memory. This method is responsible for the real job of
    fetching instructions and filling the cache.

    :param u24 addr: absolute address to read from
    :return: instruction
    :rtype: ``InstBinaryFormat_Master``
    """

    core = self.core

    inst = core.instruction_set.decode_instruction(core.memory.read_u32(addr, not_execute = False))
    opcode = inst.opcode

    if opcode not in core.instruction_set.opcode_desc_map:
      raise InvalidOpcodeError(opcode, ip = core.current_ip, core = self)

    return (inst, opcode, core.instruction_set.opcode_desc_map[opcode].execute)

class CPUDataCache(LRUCache):
  """
  Simple data cache class, based on LRU dictionary, with a limited size.
  Operates on words, and only write-back policy is supported.

  All modified entries are marked as dirty, and are NOT written back to
  memory until cache is flushed or there is a need for space and entry is
  to be removed from cache.

  Helper methods are provided, to wrap cache API to a standardized "memory
  access" API. In the future, I may extend support to more sizes, or
  restructure internal storage to keep longer blocks keyed by address (like the
  real caches do). Therefore, CPU (and others) are expected to access data
  using :py:meth:`ducky.cpu.DataCache.read_u16` and
  :py:meth:`ducky.cpu.DataCache.write_u16`, instead of accessing raw values
  using address as a dictionary key.

  Cache "owns" its entries until someone else realizes it's time to give them
  up. Then cache have to "release" entries in question - it does not have to
  write remove such entries, but it has to make sure they are consistent with
  the content of main memory.

  :param ducky.cpu.CPUCacheController controller: cache controller that will
    dispatch notifications to all caches that share this core's main memory.
  :param ducky.cpu.CPUCore core: CPU core that owns this cache.
  :param int size: maximal number of entries this cache can store.
  """

  def __init__(self, controller, core, size, *args, **kwargs):
    super(CPUDataCache, self).__init__(core.cpu.machine.LOGGER, size, *args, **kwargs)

    self.controller = controller
    self.core = core

    self.forced_writes = 0

  def __repr__(self):
    return 'dictionary-based cache, %i slots' % self.size

  def make_space(self):
    """
    Removes at least one of entries in cache, saving its content into memory
    when necessary.
    """

    addr, value = self.popitem(last = False)
    dirty, value = value

    self.core.DEBUG('%s.make_space: addr=%s, value=%s', self.__class__.__name__, ADDR_FMT(addr), UINT16_FMT(value))

    if dirty:
      self.core.memory.write_u16(addr, value)

    self.prunes += 1

  def get_object(self, addr):
    """
    Read word from memory. This method is responsible for the real job of
    fetching data and filling the cache.

    :param u24 addr: absolute address to read from.
    :returns: cache entry
    :rtype: ``list``
    """

    self.core.DEBUG('%s.get_object: address=%s', self.__class__.__name__, ADDR_FMT(addr))

    self.controller.flush_entry_references(addr, caller = self.core)

    return [False, self.core.memory.read_u16(addr)]

  def read_u8(self, addr):
    """
    Read byte from cache. Value is read from memory if it is not yet present
    in cache.

    :param u24 addr: absolute address to read from.
    :rtype: u8
    """

    word_addr = addr & ~1

    self.core.DEBUG('%s.read_u8: addr=%s, word_addr=%s', self.__class__.__name__, ADDR_FMT(addr), ADDR_FMT(word_addr))

    word = self.read_u16(word_addr)
    return (word & 0xFF) if addr == word_addr else (word & 0xFF00) >> 8

  def write_u8(self, addr, value):
    """
    Write byte to cache. Value in cache is overwritten, and marked as dirty. It
    is not written back to the memory yet.

    :param u24 addr: absolute address to modify.
    :param u8 value: new value to write.
    """

    word_addr = addr & ~1

    self.core.DEBUG('%s.write_u8: addr=%s, word_addr=%s, vaue=%s', self.__class__.__name__, ADDR_FMT(addr), ADDR_FMT(word_addr), UINT16_FMT(value))

    word = self.read_u16(word_addr)

    if addr == word_addr:
      self.write_u16(word_addr, (word & 0xFF00) | (value & 0xFF))
    else:
      self.write_u16(word_addr, (word & 0x00FF) | (value & 0xFF) << 8)

  def read_u16(self, addr):
    """
    Read word from cache. Value is read from memory if it is not yet present
    in cache.

    :param u24 addr: absolute address to read from.
    :rtype: u16
    """

    self.core.DEBUG('%s.read_u16: addr=%s', self.__class__.__name__, ADDR_FMT(addr))

    if self.core.memory.page((addr & PAGE_MASK) >> PAGE_SHIFT).cache is False:
      self.core.DEBUG('%s.read_u16: read directly from uncacheable page', self.__class__.__name__)
      return self.core.memory.read_u16(addr)

    return self[addr][1]

  def write_u16(self, addr, value):
    """
    Write word to cache. Value in cache is overwritten, and marked as dirty. It
    is not written back to the memory yet.

    :param u24 addr: absolute address to modify.
    :param u16 value: new value to write.
    """

    self.core.DEBUG('%s.write_u16: addr=%s, value=%s', self.__class__.__name__, ADDR_FMT(addr), UINT16_FMT(value))

    if self.core.memory.page((addr & PAGE_MASK) >> PAGE_SHIFT).cache is False:
      self.core.DEBUG('%s.write_u16: write directly to uncacheable page', self.__class__.__name__)
      self.core.memory.write_u16(addr, value)
      return

    if addr not in self:
      self.__missing__(addr)

    self[addr] = [True, value]

    self.controller.release_entry_references(addr, caller = self.core)

  def release_entry_references(self, addr, writeback = True, remove = True):
    """
    Give up cached entry.

    :param u24 addr: entry address.
    :param bool writeback: if ``True``, entries is written back to memory.
    :param bool remove: if ``True``, entry is removed from cache.
    """

    self.core.DEBUG('%s.release_entry_references: address=%s, writeback=%s, remove=%s', self.__class__.__name__, ADDR_FMT(addr), writeback, remove)

    if writeback:
      dirty, value = self.get(addr, (None, None))

      if dirty is None and value is None:
        self.core.DEBUG('%s.release_entry_references: not cached', self.__class__.__name__,)
        return

      if dirty:
        self.core.DEBUG('%s.release_entry_reference: write back', self.__class__.__name__)
        self.core.memory.write_u16(addr, value)

      if not remove:
        self[addr] = (False, value)
        return

    if remove:
      self.core.DEBUG('%s.release_entry_reference: remove', self.__class__.__name__)

      try:
        del self[addr]

      except KeyError:
        pass

  def release_page_references(self, page, writeback = True, remove = True):
    """
    Give up cached entries located a specific memory page.

    :param ducky.mm.MemoryPage page: referenced page.
    :param bool writeback: if ``True``, entries are written back to memory.
    :param bool remove: if ``True``, entries are removed from cache.
    """

    self.core.DEBUG('%s.release_page_references: page=%s, writeback=%s, remove=%s', self.__class__.__name__, page.index, writeback, remove)

    addresses = [i for i in range(page.base_address, page.base_address + PAGE_SIZE, 2)]

    for addr in [addr for addr in iterkeys(self) if addr in addresses]:
      self.release_entry_references(addr, writeback = writeback, remove = remove)

  def release_area_references(self, address, size, writeback = True, remove = True):
    """
    Give up cached entries located in a specific memory range.

    :param u24 address: address of the first byte of area.
    :param u24 size: length of the area in bytes.
    :param bool writeback: if ``True``, entries are written back to memory.
    :param bool remove: if ``True``, entries are removed from cache.
    """

    self.core.DEBUG('%s.remove_area_references: address=%s, size=%s, writeback=%s, remove=%s', self.__class__.__name__, ADDR_FMT(address), UINT16_FMT(size), writeback, remove)

    addresses = [i for i in range(address, address + size, 2)]

    for addr in [addr for addr in iterkeys(self) if addr in addresses]:
      self.release_entry_references(addr, writeback = writeback, remove = remove)

  def release_references(self, writeback = True, remove = True):
    """
    Give up all cached entries.

    :param boolean writeback: if ``True``, entries are written back to memory.
    :param boolean remove: if ``True``, entries are removed from cache.
    """

    self.core.DEBUG('%s.release_references: writeback=%s, remove=%s', self.__class__.__name__, writeback, remove)

    for addr in list(iterkeys(self)):
      self.release_entry_references(addr, writeback = writeback, remove = remove)

class CPUCacheController(object):
  """
  Cache controllers manages consistency and coherency of all CPU caches.
  Provides methods that informs all involved parties about invalidation
  of cache entries.

  :param ducky.machine.Machine machine: VM this controller belongs to.
  """

  def __init__(self, machine):
    self.machine = machine

    self.cores = []

  def register_core(self, core):
    """
    Register CPU core as a listener. Core's data cache will get all
    notifications about invalidated cache entries.

    :param ducky.cpu.CPUCore core: core to be registered.
    """

    self.machine.DEBUG('%s.register_core: core=%s', self.__class__.__name__, core)

    self.cores.append(core)

  def unregister_core(self, core):
    """
    Unregister CPU core. Core's data cache will no longer receive any
    notifications about invalidated cache entries.
    """

    self.machine.DEBUG('%s.unregister_core: core=%s', self.__class__.__name__, core)

    self.cores.remove(core)

  def flush_entry_references(self, address, caller = None):
    """
    Instruct caches to save a single entry back to memory.

    :param u24 address: entry address.
    :param ducky.cpu.CPUCore caller: core requesting this action. If set, all
      cores except this particular one will be instruct to save their cached
      entry.
    """

    self.machine.DEBUG('%s.flush_entry_references: caller=%s, address=%s', self.__class__.__name__, caller, ADDR_FMT(address))

    for core in [core for core in self.cores if core is not caller]:
      core.data_cache.release_entry_references(address, writeback = True, remove = False)

  def release_entry_references(self, address, caller = None):
    """
    Instruct caches to give up one cached entry.

    :param u24 address: entry address.
    :caller ducky.cpu.CPUCore caller: core requesting this action. If set, all
      cores except this particular one will be instruct to throw away cached
      entry without saving it back to memory. Otherwise, caches will save their
      version of entry before removing it.
    """

    self.machine.DEBUG('%s.release_entry_references: caller=%s, addresss=%s', self.__class__.__name__, caller, ADDR_FMT(address))

    writeback = True if caller is None else False
    for core in [core for core in self.cores if core is not caller]:
      core.data_cache.release_entry_references(address, writeback = writeback, remove = True)

  def release_page_references(self, pg, caller = None):
    """
    Instruct caches to give up entries located on one page.

    :param ducky.mm.MemoryPage pg: referenced page.
    :caller ducky.cpu.CPUCore caller: core requesting this action. If set, all
      cores except this particular one will be instruct to throw away cached
      entries without saving them back to memory. Otherwise, caches will save
      their version of entries before removing it.
    """

    self.machine.DEBUG('%s.release_page_references: caller=%s, pg=%s', self.__class__.__name__, caller, pg)

    writeback = True if caller is None else False
    for core in [core for core in self.cores if core is not caller]:
      core.data_cache.release_page_references(pg, writeback = writeback, remove = True)

  def release_area_references(self, address, size, caller = None):
    """
    Instruct caches to give up entries in memory area.

    :param u24 address: address of the first byte of area.
    :param u24 size: length of the area in bytes.
    :caller ducky.cpu.CPUCore caller: core requesting this action. If set, all
      cores except this particular one will be instruct to throw away cached
      entries without saving them back to memory. Otherwise, caches will save
      their version of entries before removing it.
    """

    self.machine.DEBUG('%s.release_area_references: caller=%s, address=%s, size=%s', self.__class__.__name__, caller, ADDR_FMT(address), ADDR_FMT(size))

    writeback = True if caller is None else False
    for core in [core for core in self.cores if core is not caller]:
      core.data_cache.release_area_references(address, size, writeback = writeback, remove = True)

  def release_references(self, caller = None):
    """
    Instruct caches to give up all cached entries.

    :param ducky.cpu.CPUCore caller: core requesting this action. If set, all
      cores except this particular one will be instruct to throw away all their
      entries without saving them back. Otherwise, caches will save their entries
      before removing them.
    """

    self.machine.DEBUG('%s.release_references: caller=%s', self.__class__.__name__, caller)

    writeback = True if caller is None else False
    for core in [core for core in self.cores if core is not caller]:
      core.data_cache.release_references(writeback = writeback, remove = True)


class CPUCore(ISnapshotable, IMachineWorker):
  """
  This class represents the main workhorse, one of CPU cores. Reads
  instructions, executes them, has registers, caches, handles interrupts,
  ...

  :param int coreid: id of this core. Usually, it's its serial number but it
    has no special meaning.
  :param ducky.cpu.CPU cpu: CPU that owns this core.
  :param ducky.mm.MemoryController memory_controller: use this controller to
    access main memory.
  """

  def __init__(self, coreid, cpu, memory_controller, cache_controller):
    super(CPUCore, self).__init__()

    config = cpu.machine.config

    self.cpuid = '#{}:#{}'.format(cpu.id, coreid)
    self.cpuid_prefix = self.cpuid + ':'

    def __log(logger, *args, **kwargs):
      args = ('%s ' + args[0],) + (self.cpuid_prefix,) + args[1:]
      logger(*args)

    def __log_exception(logger_fn, exc):
      self.cpu.machine.LOGGER.exception('Exception raised in CPU core')
      do_log_cpu_core_state(self, logger = logger_fn, disassemble = False if isinstance(exc, InvalidOpcodeError) else True)

    self.LOGGER = cpu.machine.LOGGER
    self.DEBUG = lambda *args, **kwargs: __log(self.cpu.machine.DEBUG, *args, **kwargs)
    self.INFO  = lambda *args, **kwargs: __log(self.cpu.machine.INFO, *args, **kwargs)
    self.WARN  = lambda *args, **kwargs: __log(self.cpu.machine.WARN, *args, **kwargs)
    self.ERROR = lambda *args, **kwargs: __log(self.cpu.machine.ERROR, *args, **kwargs)
    self.EXCEPTION = functools.partial(__log_exception, self.ERROR)

    self.id = coreid
    self.cpu = cpu
    self.memory = memory_controller

    self.registers = registers.RegisterSet()

    self.ivt_address = config.getint('cpu', 'ivt-address', DEFAULT_IVT_ADDRESS)

    self.instruction_set = DuckyInstructionSet
    self.instruction_set_stack = []

    self.current_ip = None
    self.current_instruction = None

    self.alive = False
    self.running = False
    self.idle = False

    self.core_profiler = profiler.STORE.get_core_profiler(self) if profiler.STORE.is_cpu_enabled() else None

    self.exit_code = 0

    self.frames = []
    self.check_frames = cpu.machine.config.getbool('cpu', 'check-frames', default = False)

    self.debug = None

    self.cache_controller = cache_controller

    self.instruction_cache = InstructionCache(self, config.getint('cpu', 'inst-cache', default = DEFAULT_CORE_INST_CACHE_SIZE))

    self.data_cache = None
    if config.getbool('cpu', 'data-cache-enabled', True):
      driver = config.get('cpu', 'data-cache-driver', 'python')

      import platform

      if platform.python_implementation() == 'PyPy':
        if driver != 'python':
          self.WARN('Running on PyPy, forcing Python data cache implementation')
          driver = 'python'

      elif platform.python_implementation() == 'CPython':
        pass

      elif driver != 'python':
        self.WARN('Running on unsupported platform, forcing Python data cache implementation')
        driver = 'python'

      if driver == 'native':
        from ..native.data_cache import CPUDataCache as DC

        self.data_cache = DC(cache_controller, self,
                             config.getint('cpu', 'data-cache-size', DEFAULT_CORE_DATA_CACHE_SIZE),
                             config.getint('cpu', 'data-cache-line', DEFAULT_CORE_DATA_CACHE_LINE_LENGTH),
                             config.getint('cpu', 'data-cache-assoc', DEFAULT_CORE_DATA_CACHE_LINE_ASSOC))

      elif driver == 'python':
        self.data_cache = CPUDataCache(cache_controller, self, config.getint('cpu', 'data-cache-size', default = DEFAULT_CORE_DATA_CACHE_SIZE))

      else:
        raise InvalidResourceError('Unknown data cache driver: driver=%s' % driver)

      self.cache_controller.register_core(self)

    if self.data_cache is not None:
      self.MEM_IN8   = self.data_cache.read_u8
      self.MEM_IN16  = self.data_cache.read_u16
      self.MEM_IN32  = self.memory.read_u32
      self.MEM_OUT8  = self.data_cache.write_u8
      self.MEM_OUT16 = self.data_cache.write_u16
      self.MEM_OUT32 = self.memory.write_u32

    else:
      self.MEM_IN8   = self.memory.read_u8
      self.MEM_IN16  = self.memory.read_u16
      self.MEM_IN32  = self.memory.read_u32
      self.MEM_OUT8  = self.memory.write_u8
      self.MEM_OUT16 = self.memory.write_u16
      self.MEM_OUT32 = self.memory.write_u32

    self.coprocessors = {}
    if self.cpu.machine.config.getbool('cpu', 'math-coprocessor', False):
      from .coprocessor import math_copro

      self.math_coprocessor = self.coprocessors['math'] = math_copro.MathCoprocessor(self)

  def has_coprocessor(self, name):
    return hasattr(self, '{}_coprocessor'.format(name))

  def __repr__(self):
    return '#{}:#{}'.format(self.cpu.id, self.id)

  def save_state(self, parent):
    state = parent.add_child('core{}'.format(self.id), CPUCoreState())

    state.cpuid = self.cpu.id
    state.coreid = self.id

    state.registers = []

    for i, reg in enumerate(REGISTER_NAMES):
      if reg == 'flags':
        value = int(self.registers.flags.to_uint16())

      else:
        value = int(self.registers.map[reg].value)

      state.registers.append(value)

    state.ivt_address = self.ivt_address

    state.exit_code = self.exit_code
    state.idle = self.idle
    state.alive = self.alive
    state.running = self.running

    if self.has_coprocessor('math'):
      self.math_coprocessor.save_state(state)

  def load_state(self, state):
    for i, reg in enumerate(REGISTER_NAMES):
      if reg == 'flags':
        self.registers.flags.load_uint16(state.registers[i])

      else:
        self.registers.map[reg].value = state.registers[i]

    self.ivt_address = state.ivt_address

    self.exit_code = state.exit_code
    self.idle = state.idle
    self.alive = state.alive
    self.running = state.running

    if self.has_coprocessor('math'):
      self.math_coprocessor.load_state(state.get_children()['math_coprocessor'])

  def init_debug_set(self):
    if self.debug is None:
      from .. import debugging
      self.debug = debugging.DebuggingSet(self)

  def FLAGS(self):
    return self.registers.flags

  def REG(self, reg):
    return self.registers.map[reg]

  def IP(self):
    return self.registers.ip

  def SP(self):
    return self.registers.sp

  def FP(self):
    return self.registers.fp

  def CS(self):
    return self.registers.cs

  def DS(self):
    return self.registers.ds

  def CS_ADDR(self, address):
    return (self.registers.cs.value & 0xFF) * SEGMENT_SIZE * PAGE_SIZE + address

  def DS_ADDR(self, address):
    return (self.registers.ds.value & 0xFF) * SEGMENT_SIZE * PAGE_SIZE + address

  def reset(self, new_ip = 0):
    """
    Reset core's state. All registers are set to zero, all flags are set to zero,
    except ``HWINT`` flag which is set to one, and ``IP`` is set to requested value.
    Both instruction and data cached are flushed.

    :param u16 new_ip: new ``IP`` value, defaults to zero
    """

    for reg in registers.RESETABLE_REGISTERS:
      self.REG(reg).value = 0

    self.registers.flags.privileged = 0
    self.registers.flags.hwint = 1
    self.registers.flags.e = 0
    self.registers.flags.z = 0
    self.registers.flags.o = 0
    self.registers.flags.s = 0

    self.registers.ip.value = new_ip

    self.instruction_cache.clear()
    if self.data_cache:
      self.data_cache.clear()

  def __symbol_for_ip(self):
    ip = self.registers.ip

    symbol, offset = self.cpu.machine.get_symbol_by_addr(self.registers.cs.value, ip.value)

    if not symbol:
      self.WARN('symbol_for_ip: Unknown jump target: %s', ip)
      return

    self.DEBUG('symbol_for_ip: %s%s (%s)', symbol, ' + {}'.format(offset if offset != 0 else ''), ip.value)

  def backtrace(self):
    bt = []

    if self.check_frames:
      for frame in self.frames:
        symbol, offset = self.cpu.machine.get_symbol_by_addr(frame.CS, frame.IP)
        bt.append((frame.IP, symbol, offset))

      return bt

    bt = []

    for frame_index, frame in enumerate(self.frames):
      ip = self.memory.read_u16(frame.address + 2, privileged = True)
      symbol, offset = self.cpu.machine.get_symbol_by_addr(frame.CS, ip)

      bt.append((ip, symbol, offset))

    ip = self.registers.ip.value - 4
    symbol, offset = self.cpu.machine.get_symbol_by_addr(self.registers.cs.value, ip)
    bt.append((ip, symbol, offset))

    return bt

  def raw_push(self, val):
    """
    Push value on stack. ``SP`` is decremented by two, and value is written at this new address.

    :param u16 val: value to be pushed
    """

    self.DEBUG("raw_push: sp=%s, ds-scp=%s, value=%s", ADDR_FMT(self.registers.sp.value), ADDR_FMT(self.DS_ADDR(self.registers.sp.value)), UINT16_FMT(val))

    self.registers.sp.value -= 2
    self.MEM_OUT16(self.DS_ADDR(self.registers.sp.value), val)

  def raw_pop(self):
    """
    Pop value from stack. 2 byte number is read from address in ``SP``, then ``SP`` is incremented by two.

    :return: popped value
    :rtype: ``u16``
    """

    ret = self.MEM_IN16(self.DS_ADDR(self.registers.sp.value))
    self.registers.sp.value += 2
    return ret

  def push(self, *regs):
    for reg_id in regs:
      reg = self.registers.map[reg_id]

      self.DEBUG('push: %s (%s) at %s', reg_id, reg.to_string() if isinstance(reg, FlagsRegister) else UINT16_FMT(reg.value), UINT16_FMT(self.registers.sp.value - 2))
      value = self.registers.flags.to_uint16() if reg_id == Registers.FLAGS else self.registers.map[reg_id].value
      self.raw_push(value)

  def pop(self, *regs):
    for reg_id in regs:
      if reg_id == Registers.FLAGS:
        reg = self.registers.flags
        reg.load_uint16(self.raw_pop())
      else:
        reg = self.registers.map[reg_id]
        reg.value = self.raw_pop()

      self.DEBUG('pop: %s (%s) from %s', reg_id, reg.to_string() if isinstance(reg, FlagsRegister) else UINT16_FMT(reg.value), UINT16_FMT(self.registers.sp.value - 2))

  def create_frame(self):
    """
    Create new call stack frame. Push ``IP`` and ``FP`` registers and set ``FP`` value to ``SP``.
    """

    self.DEBUG('create_frame')

    self.push(Registers.IP, Registers.FP)

    self.registers.fp.value = self.registers.sp.value

    if self.check_frames:
      self.frames.append(StackFrame(self.registers.cs.value, self.registers.ds.value, self.registers.fp.value))

  def destroy_frame(self):
    """
    Destroy current call frame. Pop ``FP`` and ``IP`` from stack, by popping ``FP`` restores previous frame.

    :raises CPUException: if current frame does not match last created frame.
    """

    self.DEBUG('destroy_frame')

    if self.check_frames:
      if self.frames[-1].FP != self.registers.sp.value:
        raise CPUException('Leaving frame with wrong SP: IP={}, saved SP={}, current SP={}'.format(ADDR_FMT(self.registers.ip.value), ADDR_FMT(self.frames[-1].FP), ADDR_FMT(self.registers.sp.value)))

      self.frames.pop()

    self.pop(Registers.FP, Registers.IP)

    self.__symbol_for_ip()

  def __load_interrupt_vector(self, index):
    self.DEBUG('load_interrupt_vector: ivt=%s, index=%i', ADDR_FMT(self.ivt_address), index)

    desc = InterruptVector()

    vector_address = self.ivt_address + index * InterruptVector.SIZE

    desc.cs = self.MEM_IN8(vector_address)
    desc.ds = self.MEM_IN8(vector_address + 1)
    desc.ip = self.MEM_IN16(vector_address + 2)

    return desc

  def __enter_interrupt(self, index):
    """
    Prepare CPU for handling interrupt routine. New stack is allocated, content fo registers
    is saved onto this new stack, and new call frame is created on this stack. CPU is switched
    into privileged mode. ``CS`` and ``IP`` are set to values, stored in interrupt descriptor
    table at specified offset.

    :param u24 table_address: address of interrupt descriptor table
    :param int index: interrupt number, its index into IDS
    """

    self.DEBUG('__enter_interrupt: index=%i', index)

    iv = self.__load_interrupt_vector(index)

    stack_pg, sp = self.memory.alloc_stack(segment = iv.ds)

    old_SP = self.registers.sp.value
    old_DS = self.registers.ds.value

    self.registers.ds.value = iv.ds
    self.registers.sp.value = sp

    self.raw_push(old_DS)
    self.raw_push(old_SP)
    self.push(Registers.CS, Registers.FLAGS)
    self.create_frame()

    self.privileged = True

    self.registers.cs.value = iv.cs
    self.registers.ip.value = iv.ip

    if self.check_frames:
      self.frames[-1].IP = iv.ip

    self.instruction_set_stack.append(self.instruction_set)
    self.instruction_set = DuckyInstructionSet

  def exit_interrupt(self):
    """
    Restore CPU state after running a interrupt routine. Call frame is destroyed, registers
    are restored, stack is returned back to memory pool.
    """

    self.DEBUG('exit_interrupt')

    self.destroy_frame()
    self.pop(Registers.FLAGS, Registers.CS)

    stack_page = self.memory.page(addr_to_page(self.DS_ADDR(self.registers.sp.value)))

    old_SP = self.raw_pop()
    old_DS = self.raw_pop()

    self.registers.ds.value = old_DS
    self.registers.sp.value = old_SP

    if self.data_cache:
      self.data_cache.release_page_references(stack_page, writeback = False)
    self.memory.free_page(stack_page)

    self.instruction_set = self.instruction_set_stack.pop(0)

  def do_int(self, index):
    """
    Handle software interrupt. Real software interrupts cause CPU state to be saved
    and new stack and register values are prepared by ``__enter_interrupt`` method,
    virtual interrupts are simply triggered without any prior changes of CPU state.

    :param int index: interrupt number
    """

    self.DEBUG('do_int: %s', index)

    if index in self.cpu.machine.virtual_interrupts:
      self.DEBUG('do_int: calling virtual interrupt')

      self.cpu.machine.virtual_interrupts[index].run(self)

      self.DEBUG('do_int: virtual interrupt finished')

    else:
      self.__enter_interrupt(index)

      self.DEBUG('do_int: CPU state prepared to handle interrupt')

  def __do_irq(self, index):
    """
    Handle hardware interrupt. CPU state is saved and prepared for interrupt routine
    by calling ``__enter_interrupt`` method. Receiving of next another interrupts
    is prevented by clearing ``HWINT`` flag, and ``idle`` flag is set to ``False``.
    """

    self.DEBUG('__do_irq: %s', index)

    self.__enter_interrupt(index)
    self.registers.flags.hwint = 0
    self.change_runnable_state(idle = False)

    self.DEBUG('__do_irq: CPU state prepared to handle IRQ')
    log_cpu_core_state(self)

  def irq(self, index):
    try:
      self.__do_irq(index)

    except (CPUException, ZeroDivisionError, AccessViolationError) as e:
      e.exc_stack = sys.exc_info()
      self.die(e)

  # Do it this way to avoid pylint' confusion
  def __get_privileged(self):
    return self.registers.flags.privileged == 1

  def __set_privileged(self, value):
    self.registers.flags.privileged = 1 if value is True else 0

  privileged = property(__get_privileged, __set_privileged)

  def check_protected_ins(self):
    """
    Raise ``AccessViolationError`` if core is not running in privileged mode.

    This method should be used by instruction handlers that require privileged mode, e.g. protected instructions.

    :raises AccessViolationError: if the core is not in privileged mode
    """

    if not self.privileged:
      raise AccessViolationError('Instruction not allowed in unprivileged mode: inst={}'.format(self.current_instruction))

  def check_protected_reg(self, reg):
    if self.privileged:
      return

    if reg in registers.PROTECTED_REGISTERS:
      raise AccessViolationError('Access not allowed in unprivileged mode: inst={}, reg={}'.format(self.current_instruction, reg))

  def check_protected_port(self, port):
    if port not in self.cpu.machine.ports:
      raise InvalidResourceError('Unhandled port: port={}'.format(UINT16_FMT(port)))

    if self.privileged:
      return

    if self.cpu.machine.ports[port].is_port_protected(port):
      raise AccessViolationError('Access to port not allowed in unprivileged mode: inst={}, port={}'.format(self.current_instruction, port))

  def update_arith_flags(self, reg):
    """
    Set relevant arithmetic flags according to content of registers. Flags are set to zero at the beginning,
    then content of each register is examined, and ``S`` and ``Z`` flags are set.

    ``E`` flag is not touched, ``O`` flag is set to zero.

    :param c_types.c_ushort reg: ``u16`` bit-wide register
    """

    F = self.registers.flags

    F.z = 0
    F.o = 0
    F.s = 0

    if reg.value == 0:
      F.z = 1

    if reg.value & 0x8000 != 0:
      F.s = 1

  def RI_VAL(self, inst):
    return self.registers.map[inst.ireg].value if inst.is_reg == 1 else inst.immediate

  def JUMP(self, inst):
    """
    Change execution flow by modifying IP. Signals profiler that jump was executed.

    :param inst: instruction that caused jump
    """

    if inst.is_reg == 1:
      self.registers.ip.value = self.registers.map[inst.ireg].value
    else:
      self.registers.ip.value += inst.immediate

    self.__symbol_for_ip()

  def CMP(self, x, y, signed = True):
    """
    Compare two numbers, and update relevant flags. Signed comparison is used unless ``signed`` is ``False``.
    All arithmetic flags are set to zero before the relevant ones are set.

    ``O`` flag is reset like the others, therefore caller has to take care of it's setting if it's required
    to set it.

    :param u16 x: left hand number
    :param u16 y: right hand number
    :param bool signed: use signed, defaults to ``True``
    """

    F = self.registers.flags

    F.e = 0
    F.z = 0
    F.o = 0
    F.s = 0

    if x == y:
      F.e = 1

      if x == 0:
        F.z = 1

      return

    if signed:
      x = i16(x).value
      y = i16(y).value

    if x < y:
      F.s = 1

    elif x > y:
      F.s = 0

  def OFFSET_ADDR(self, inst):
    self.DEBUG('offset addr: inst=%s', inst)

    base = self.registers.map[inst.areg].value
    offset = self.registers.map[inst.oreg].value if inst.is_reg == 1 else inst.immediate

    if inst.is_segment == 1:
      addr = (base & 0xFF) * SEGMENT_SIZE * PAGE_SIZE + offset

    else:
      addr = self.DS_ADDR(base + offset)

    self.DEBUG('offset addr: base=%s, offset=%s, addr=%s', ADDR_FMT(base), ADDR_FMT(offset), ADDR_FMT(addr))
    return addr

  def step(self):
    """
    Perform one "step" - fetch next instruction, increment IP, and execute instruction's code (see inst_* methods)
    """

    self.DEBUG('----- * ----- * ----- * ----- * ----- * ----- * ----- * -----')

    if self.debug is not None:
      self.debug.enter_step()

      if not self.running:
        return

    # Read next instruction
    self.DEBUG('"FETCH" phase')

    ip = self.registers.ip
    self.current_ip = ip.value

    self.DEBUG('fetch instruction: cs=%s, ip=%s', UINT8_FMT(self.registers.cs.value), ADDR_FMT(ip.value))

    try:
      self.current_instruction, opcode, execute = self.instruction_cache[self.CS_ADDR(ip.value)]
      ip.value += 4

      self.DEBUG('"EXECUTE" phase: %s %s', UINT16_FMT(self.current_ip), self.instruction_set.disassemble_instruction(self.current_instruction))
      log_cpu_core_state(self)

      execute(self, self.current_instruction)

    except (InvalidOpcodeError, AccessViolationError, InvalidResourceError) as e:
      self.die(e)
      return

    cnt = self.registers.cnt
    cnt.value += 1

    self.DEBUG('"SYNC" phase:')
    log_cpu_core_state(self)

    if self.core_profiler is not None:
      self.core_profiler.take_sample()

    if self.debug is not None:
      self.debug.exit_step()

  def change_runnable_state(self, alive = None, running = None, idle = None):
    old_state = self.alive and self.running and not self.idle

    if alive is not None:
      self.alive = alive

    if running is not None:
      self.running = running

    if idle is not None:
      self.idle = idle

    new_state = self.alive and self.running and not self.idle

    if old_state != new_state:
      if new_state is True:
        self.cpu.machine.reactor.task_runnable(self)

      else:
        self.cpu.machine.reactor.task_suspended(self)

  def suspend(self):
    self.DEBUG('CPUCore.suspend')

    self.change_runnable_state(running = False)
    self.cpu.core_suspended()

  def wake_up(self):
    self.DEBUG('CPUCore.wake_up')

    self.change_runnable_state(running = True)
    self.cpu.core_running()

  def die(self, exc):
    self.DEBUG('CPUCore.die')

    self.exit_code = 1

    self.EXCEPTION(exc)

    self.halt()

  def halt(self):
    self.DEBUG('CPUCore.halt')

    if self.data_cache is not None:
      self.data_cache.release_references()
      self.cpu.machine.cpu_cache_controller.unregister_core(self)

    self.change_runnable_state(alive = False, running = False)
    self.cpu.core_suspended()
    self.cpu.core_halted()

    log_cpu_core_state(self)

    self.cpu.machine.reactor.remove_task(self)

    self.INFO('CPU core halted')

  def run(self):
    try:
      self.step()

    except (CPUException, ZeroDivisionError, AccessViolationError) as e:
      e.exc_stack = sys.exc_info()
      self.die(e)

    except Exception as e:
      e.exc_stack = sys.exc_info()
      self.die(e)

  def boot(self, init_state):
    self.DEBUG('CPUCore.boot')

    self.reset()

    cs, ds, sp, ip, privileged = init_state

    self.registers.cs.value = cs
    self.registers.ds.value = ds
    self.registers.ip.value = ip
    self.registers.sp.value = sp
    self.registers.fp.value = sp
    self.registers.flags.privileged = 1 if privileged else 0

    log_cpu_core_state(self)

    self.cpu.machine.reactor.add_task(self)
    self.change_runnable_state(alive = True, running = True)
    self.cpu.core_alive()
    self.cpu.core_running()

    if self.core_profiler is not None:
      self.core_profiler.enable()

    self.INFO('CPU core is up')
    if self.data_cache is not None:
      self.INFO('  {}'.format(repr(self.data_cache)))
    if self.instruction_cache is not None:
      self.INFO('  {} IC slots'.format(self.instruction_cache.size))
    self.INFO('  check-frames: %s', 'yes' if self.check_frames else 'no')
    if self.coprocessors:
      self.INFO('  coprocessor: %s', ' '.join(sorted(iterkeys(self.coprocessors))))

class CPU(ISnapshotable, IMachineWorker):
  def __init__(self, machine, cpuid, memory_controller, cache_controller, cores = 1):
    super(CPU, self).__init__()

    self.cpuid_prefix = '#{}:'.format(cpuid)

    def __log(logger, *args):
      args = ('%s ' + args[0],) + (self.cpuid_prefix,) + args[1:]
      logger(*args)

    self.DEBUG = lambda *args: __log(self.machine.DEBUG, *args)
    self.INFO  = lambda *args: __log(self.machine.INFO, *args)
    self.WARN  = lambda *args: __log(self.machine.WARN, *args)
    self.ERROR = lambda *args: __log(self.machine.ERROR, *args)
    self.EXCEPTION = lambda *args: __log(self.machine.EXCEPTION, *args)

    self.machine = machine
    self.id = cpuid

    self.memory = memory_controller
    self.cache_controller = cache_controller

    self.cores = []
    for i in range(0, cores):
      __core = CPUCore(i, self, self.memory, self.cache_controller)
      self.cores.append(__core)

    self.cnt_living_cores = 0
    self.cnt_running_cores = 0

    self.machine.console.register_commands([
      ('sc', cmd_set_core),
      ('st', cmd_core_state),
      ('cont', cmd_cont)
    ])

  def save_state(self, parent):
    state = parent.add_child('cpu{}'.format(self.id), CPUState())

    for core in self.cores:
      core.save_state(state)

  def load_state(self, state):
    for core_state in itervalues(state.get_children()):
      self.cores[core_state.coreid].load_state(core_state)

  def core_alive(self):
    """
    Signal CPU that one of cores is now alive.
    """

    self.cnt_living_cores += 1
    self.machine.core_alive()

  def core_halted(self):
    """
    Signal CPU that one of cores is no longer alive.
    """

    self.cnt_living_cores -= 1
    self.machine.core_halted()

  def core_running(self):
    """
    Signal CPU that one of cores is now running.
    """

    self.cnt_running_cores += 1

  def core_suspended(self):
    """
    Signal CPU that one of cores is now suspended.
    """

    self.cnt_running_cores -= 1

  @property
  def living_cores(self):
    return [__core for __core in self.cores if __core.alive is True]

  @property
  def halted_cores(self):
    return [__core for __core in self.cores if __core.alive is not True]

  @property
  def running_cores(self):
    return [__core for __core in self.cores if __core.running is True]

  @property
  def suspended_cores(self):
    return [__core for __core in self.cores if __core.running is not True]

  def suspend(self):
    self.DEBUG('CPU.suspend')

    for core in self.running_cores:
      core.suspend()

  def wake_up(self):
    self.DEBUG('CPU.wake_up')

    for core in self.suspended_cores:
      core.wake_up()

  def die(self, exc):
    self.DEBUG('CPU.die')

    self.EXCEPTION(exc)

    self.halt()

  def halt(self):
    self.DEBUG('CPU.halt')

    for core in self.living_cores:
      core.halt()

    self.INFO('CPU halted')

  def boot(self, init_states):
    self.DEBUG('CPU.boot')

    for core in self.cores:
      if init_states:
        core.boot(init_states.pop(0))

    self.INFO('CPU is up')

def cmd_set_core(console, cmd):
  """
  Set core address of default core used by control commands: sc <coreid>
  """

  M = console.master.machine

  try:
    core = M.core(cmd[1])

  except InvalidResourceError:
    console.writeln('go away')
    return

  console.default_core = core

  console.writeln('# OK: default core is %s', core.cpuid)

def cmd_cont(console, cmd):
  """
  Continue execution until next breakpoint is reached: cont
  """

  if console.default_core is None:
    console.writeln('# ERR: no core selected')
    return

  if console.default_core.running:
    console.writeln('# ERR: core is not suspended')
    return

  console.default_core.wake_up()

  console.writeln('# OK')

def cmd_step(console, cmd):
  """
  Step one instruction forward
  """

  if console.default_core is None:
    return

  if console.default_core.running:
    return

  console.default_core.run()

def cmd_next(console, cmd):
  """
  Proceed to the next instruction in the same stack frame.
  """

  core = console.default_core if hasattr(console, 'default_core') else console.machine.cpus[0].cores[0]

  if not core.is_suspended():
    return

  def __ip_addr(offset = 0):
    return core.CS_ADDR(core.registers.ip.value + offset)

  try:
    inst = core.instruction_set.decode_instruction(core.memory.read_u32(__ip_addr()))

    if inst.opcode == core.instruction_set.opcodes.CALL:
      from ..debugging import add_breakpoint

      add_breakpoint(core, core.registers.ip.value + 4, ephemeral = True)

      core.wake_up()

    else:
      core.step()
      core.check_for_events()

      log_cpu_core_state(core, logger = core.INFO)

  except CPUException as e:
    core.die(e)

def cmd_core_state(console, cmd):
  """
  Print core state
  """

  M = console.master.machine
  core = console.default_core if console.default_core is not None else M.cpus[0].cores[0]

  do_log_cpu_core_state(core, logger = functools.partial(console.log, core.INFO))

def cmd_bt(console, cmd):
  """
  Print current backtrace
  """

  M = console.master.machine
  core = console.default_core if console.default_core is not None else M.cpus[0].cores[0]

  table = [
    ['Index', 'symbol', 'offset', 'ip']
  ]

  for index, (ip, symbol, offset) in enumerate(core.backtrace()):
    table.append([index, symbol, UINT16_FMT(offset), ADDR_FMT(ip)])

  console.table(table)



import os

from . import Device
from ..streams import InputStream, OutputStream

class Terminal(Device):
  def __init__(self, machine, name, *args, **kwargs):
    super(Terminal, self).__init__(machine, 'terminal', name, *args, **kwargs)

    self.input = None
    self.output = None

class StreamIOTerminal(Terminal):
  def __init__(self, machine, name, input = None, output = None, streams_in = None, stream_out = None, *args, **kwargs):
    super(StreamIOTerminal, self).__init__(machine, name, *args, **kwargs)

    self.input = input
    self.output = output

    self.input.master = self
    self.output.master = self

    streams_in = streams_in or []

    self.machine.DEBUG('streams_in=%s', streams_in)
    for stream in streams_in:
      self.input.enqueue_input(stream)

    self.output.set_output(stream_out)

  @staticmethod
  def get_slave_devices(machine, config, section):
    input_name = config.get(section, 'input', None)
    input_device = machine.get_device_by_name(input_name)

    if not input_name or not input_device:
      machine.ERROR('Unknown slave device %s', input_name)

    output_name = config.get(section, 'output', None)
    output_device = machine.get_device_by_name(output_name)

    if not output_name or not output_device:
      machine.ERROR('Unknown slave device %s', output_name)

    return (input_device, output_device)

  @staticmethod
  def create_from_config(machine, config, section):
    input_device, output_device = StreamIOTerminal.get_slave_devices(machine, config, section)

    streams_in = config.get(section, 'streams_in', None)
    if streams_in is not None:
      streams_in = [InputStream.create(machine.LOGGER, e.strip()) for e in streams_in.split(',')]

    return StreamIOTerminal(machine, section, input = input_device, output = output_device, streams_in = streams_in, stream_out = OutputStream.create(machine.LOGGER, config.get(section, 'stream_out', None)))

  def boot(self):
    self.machine.DEBUG('StreamIOTerminal.boot')

    super(StreamIOTerminal, self).boot()

    self.input.boot()
    self.output.boot()

    self.machine.INFO('hid: basic terminal (%s, %s)', self.input.name, self.output.name)

  def halt(self):
    self.machine.DEBUG('StreamIOTerminal.halt')

    super(StreamIOTerminal, self).halt()

    self.input.halt()
    self.output.halt()

    self.machine.DEBUG('Standard terminal halted.')


class StandardIOTerminal(StreamIOTerminal):
  @staticmethod
  def create_from_config(machine, config, section):
    input_device, output_device = StreamIOTerminal.get_slave_devices(machine, config, section)

    return StandardIOTerminal(machine, section, input = input_device, output = output_device, streams_in = [InputStream.create(machine.LOGGER, '<stdin>')], stream_out = OutputStream.create(machine.LOGGER, '<stdout>'))


class StandalonePTYTerminal(StreamIOTerminal):
  def __init__(self, *args, **kwargs):
    super(StandalonePTYTerminal, self).__init__(*args, **kwargs)

    self.pttys = None

  @staticmethod
  def create_from_config(machine, config, section):
    input_device, output_device = StreamIOTerminal.get_slave_devices(machine, config, section)

    return StandalonePTYTerminal(machine, section, input = input_device, output = output_device)

  def boot(self):
    self.machine.DEBUG('StandalonePTYTerminal.boot')

    Terminal.boot(self)

    if self.pttys is not None:
      return

    import pty

    pttys = pty.openpty()

    self.machine.DEBUG('  set I/O stream: pttys=%s', pttys)

    self.input.enqueue_input(InputStream.create(self.machine.LOGGER, pttys[0]))
    self.output.set_output(OutputStream.create(self.machine.LOGGER, pttys[0]))

    self.terminal_device = os.ttyname(pttys[1]) if pttys else '/dev/unknown'

    self.input.boot()
    self.output.boot()

    self.pttys = pttys

    self.machine.INFO('hid: pty terminal (%s, %s), dev %s', self.input.name, self.output.name, self.terminal_device)

  def halt(self):
    self.machine.DEBUG('StandalonePTYTerminal.halt')

    Terminal.halt(self)

    if self.pttys is None:
      return

    self.input.halt()
    self.output.halt()

    try:
      os.close(self.pttys[1])
      os.close(self.pttys[0])

      self.pttys = None
      self.terminal_device = None

    except Exception:
      self.machine.EXCEPTION('Exception raised while closing PTY')

    self.machine.DEBUG('StandalonePTYTerminal: halted')



"""
Streams represent basic IO objects, used by devices for reading or writing
streams of data.

``Stream`` object encapsulates an actual IO object - either ``file``-like
stream, or raw file descriptor. ``Stream`` then provides basic IO methods
for accessing content of IO object, shielding user from Python2/Python3
and ``file``/descriptor differencies.
"""

import abc
import io
import os
import sys

from six import PY2, integer_types, string_types, add_metaclass

from .errors import InvalidResourceError
from .util import isfile

@add_metaclass(abc.ABCMeta)
class Stream(object):
  """
  Abstract base class of all streams.

  :param logger: logger object used for logging.
  :param desc: description of stream. This is a short, string representation
    of the stream.
  :param stream: ``file``-like stream that provides IO method (``read()`` or
    ``write``). If it is set, it is preferred IO object.
  :param int fd: raw file descriptor. ``stream`` takes precedence, otherwise this
    file descriptor is used.
  :param bool close: if ``True``, and if ``stream`` has a ``close()`` method, stream
    will provide ``close()`` method that will close the underlaying ``file``-like
    object.
  """

  def __init__(self, logger, desc, stream = None, fd = None, close = True):
    if stream is None and fd is None:
      raise InvalidResourceError('Stream "%s" must have stream object or raw file descriptor.' % desc)

    self.logger = logger
    self.DEBUG = logger.debug
    self.DEBUG = logger.info

    self.desc = desc
    self.stream = stream
    self.fd = fd

    self._raw_read = self._raw_read_stream if stream is not None else self._raw_read_fd
    self._raw_write = self._raw_write_stream if stream is not None else self._raw_write_fd

    self.close = stream.close if close and stream is not None and hasattr(stream, 'close') else self._close_dummy

  def __repr__(self):
    return '<%s %s>' % (self.__class__.__name__, self.desc)

  def has_fd(self):
    """
    Check if stream has raw file descriptor. File descriptors can be checked
    for IO availability by reactor's polling task.

    :rtype: bool
    :returns: ``True`` when stream has file descriptor.
    """

    return self.fd is not None

  def _raw_read_stream(self, size = None):
    self.DEBUG('%s._raw_read_stream: size=%s', self.__class__.__name__, size)

    size = size or 0
    return self.stream.read(size)

  def _raw_read_fd(self, size = None):
    self.DEBUG('%s._raw_read_fd: size=%s', self.__class__.__name__, size)

    size = size or io.DEFAULT_BUFFER_SIZE
    return os.read(self.fd, size)

  def _raw_write_stream(self, data):
    self.DEBUG('%s._raw_write_stream: data="%s", len=%s, type=%s', self.__class__.__name__, data, len(data), type(data))

    self.stream.write(data)

  def _raw_write_fd(self, data):
    self.DEBUG('%s._raw_write_fd: data="%s", len=%s, type=%s', self.__class__.__name__, data, len(data), type(data))

    os.write(self.fd, data)

  @abc.abstractmethod
  def read(self, size = None):
    """
    Read data from stream.

    :param int size: if set, read at maximum ``size`` bytes.
    :rtype: ``bytearray`` (Python2), ``bytes`` (Python3)
    :returns: read data, of maximum lenght of ``size``.
    """

    raise NotImplementedError('%s does not implement read method' % self.__class__.__name__)

  @abc.abstractmethod
  def write(self, buff):
    """
    Write data to stream.

    :param bytearray buff: data to write. ``bytearray`` (Python2), ``bytes`` (Python3)
    """

    raise NotImplementedError('%s does not implement write method' % self.__class__.__name__)

  def _close_dummy(self):
    """
    Dummy close method, used when stream is not expected to close underlying IO object.
    """

    self.DEBUG('%s.close: not supported', self.__class__.__name__)
    pass

class InputStream(Stream):
  if PY2:
    def read(self, size = None):
      self.DEBUG('%s.read: size=%s', self.__class__.__name__, size)

      buff = self._raw_read(size = size)
      if not buff:
        return bytearray([])

      return bytearray([ord(c) for c in buff])

  else:
    def read(self, size = None):
      self.DEBUG('%s.read: size=%s', self.__class__.__name__, size)

      buff = self._raw_read(size = size)

      return buff

  def write(self, b):
    raise NotImplementedError('%s does not implement write method' % self.__class__.__name__)

  @staticmethod
  def create(logger, desc):
    logger.debug('InputStream.create: desc=%s', desc)

    if isfile(desc):
      return FileInputStream(logger, desc)

    if hasattr(desc, 'read'):
      return MethodInputStream(logger, desc)

    if hasattr(desc, 'fileno'):
      return FDInputStream(logger, desc.fileno())

    if isinstance(desc, integer_types):
      return FDInputStream(logger, desc)

    if desc == '<stdin>':
      return StdinStream(logger)

    if isinstance(desc, string_types):
      return FileInputStream(logger, open(desc, 'rb'))

    raise InvalidResourceError('Unknown stream description: desc=%s' % desc)

class FileInputStream(InputStream):
  def __init__(self, logger, f, **kwargs):
    super(FileInputStream, self).__init__(logger, '<file %s>' % f.name, stream = f, fd = f.fileno())

class MethodInputStream(InputStream):
  def __init__(self, logger, desc, **kwargs):
    super(MethodInputStream, self).__init__(logger, repr(desc), stream = desc)

class FDInputStream(InputStream):
  def __init__(self, logger, fd, **kwargs):
    super(FDInputStream, self).__init__(logger, '<fd %s>' % fd, fd = fd)

class StdinStream(InputStream):
  def __init__(self, logger, **kwargs):
    stream = sys.stdin.buffer if hasattr(sys.stdin, 'buffer') else sys.stdin
    fd = sys.stdin.fileno() if hasattr(sys.stdin, 'fileno') else None

    super(StdinStream, self).__init__(logger, '<stdin>', stream = stream, fd = fd, close = False, **kwargs)

class OutputStream(Stream):
  def read(self, size = None):
    raise NotImplementedError('%s does not implement read method' % self.__class__.__name__)

  if PY2:
    def write(self, buff):
      self.DEBUG('%s.write: buff=%s', self.__class__.__name__, buff)

      self._raw_write(''.join([chr(b) for b in buff]))

  else:
    def write(self, buff):
      self.DEBUG('%s.write: buff=%s', self.__class__.__name__, buff)

      self._raw_write(bytes(buff))

  @staticmethod
  def create(logger, desc):
    logger.debug('OutputStream.create: desc=%s', desc)

    if isfile(desc):
      return FileOutputStream(logger, desc)

    if hasattr(desc, 'write'):
      return MethodOutputStream(logger, desc)

    if hasattr(desc, 'fileno'):
      return FDOutputStream(logger, desc.fileno())

    if isinstance(desc, integer_types):
      return FDOutputStream(logger, desc)

    if desc == '<stdout>':
      return StdoutStream(logger)

    if desc == '<stderr>':
      return StderrStream(logger)

    if isinstance(desc, string_types):
      return FileOutputStream(logger, open(desc, 'wb'))

    raise InvalidResourceError('Unknown stream description: desc=%s' % desc)

class FileOutputStream(OutputStream):
  def __init__(self, logger, f, **kwargs):
    super(FileOutputStream, self).__init__(logger, '<file %s>' % f.name, stream = f, fd = f.fileno())

class FDOutputStream(OutputStream):
  def __init__(self, logger, fd, **kwargs):
    super(FDOutputStream, self).__init__(logger, '<fd %s>' % fd, fd = fd)

class MethodOutputStream(OutputStream):
  def __init__(self, logger, desc, **kwargs):
    super(MethodOutputStream, self).__init__(logger, repr(desc), stream = desc)

class StdoutStream(OutputStream):
  def __init__(self, logger):
    stream = sys.stdout.buffer if hasattr(sys.stdout, 'buffer') else sys.stdout
    fd = sys.stdout.fileno() if hasattr(sys.stdout, 'fileno') else None

    super(StdoutStream, self).__init__(logger, '<stdout>', stream = stream, fd = fd, close = False)

class StderrStream(OutputStream):
  def __init__(self, logger):
    stream = sys.stderr.buffer if hasattr(sys.stderr, 'buffer') else sys.stderr
    fd = sys.stderr.fileno() if hasattr(sys.stderr, 'fileno') else None

    super(StderrStream, self).__init__(logger, '<stderr>', stream = stream, fd = fd, close = False)



import enum
import logging

from six import itervalues

from .interfaces import IVirtualInterrupt
from .devices import IRQList, VIRTUAL_INTERRUPTS
from .util import str2int, F
from .errors import InvalidResourceError

class Point(object):
  point_id = 0
  points = {}

  def __init__(self, debugging_set, active = True, countdown = 0):
    super(Point, self).__init__()

    self.id = Point.point_id
    Point.point_id += 1
    Point.points[self.id] = self

    self.active = active
    self.countdown = countdown
    self.debugging_set = debugging_set

  def is_triggered(self, core):
    raise NotImplementedError()

  def destroy(self):
    del Point.points[self.id]

  def __repr__(self):
    return '<%s: id=%i>' % (self.__class__.__name__, self.id)

class BreakPoint(Point):
  def __init__(self, debugging_set, ip, *args, **kwargs):
    super(BreakPoint, self).__init__(debugging_set, *args, **kwargs)

    self.ip = ip

  def is_triggered(self, core):
    core.DEBUG('core IP=%s, self IP=%s', core.IP().value, self.ip)

    return core.IP().value == self.ip

  def __repr__(self):
    return F('<BreakPoint: id={id:d}, ip={ip:W}>', id = self.id, ip = self.ip)

  @staticmethod
  def create_from_config(debugging_set, config, section):
    _get, _getbool, _getint = config.create_getters(section)

    return BreakPoint(debugging_set, _getint('address'), active = _getbool('active', True), countdown = _getint('countdown', 0))

class MemoryWatchPoint(Point):
  def __init__(self, debugging_set, address, access, *args, **kwargs):
    super(MemoryWatchPoint, self).__init__(debugging_set, *args, **kwargs)

    self.address = address
    self.access = access

  def is_triggered(self, core):
    return False

  @staticmethod
  def create_from_config(debugging_set, config, section):
    _get, _getbool, _getint = config.create_getters(section)

    return MemoryWatchPoint(debugging_set, _getint('address'), _get('access'), active = _getbool('active', True), countdown = _getint('countdown', 0))

class DebuggingSet(object):
  def __init__(self, core):
    super(DebuggingSet, self).__init__()

    self.core = core
    self.points = []

    self.triggered_points = []

    C = core.cpu.machine.console

    console_commands = [
      ('bp-list', cmd_bp_list),
      ('bp-break', cmd_bp_add_breakpoint),
      ('bp-mwatch', cmd_bp_add_memory_watchpoint),
      ('bp-active', cmd_bp_active)
    ]

    for name, handler in console_commands:
      if C.is_registered_command(name):
        continue

      C.register_command(name, handler)

  def add_point(self, p):
    self.points.append(p)

  def remove_point(self, p):
    self.points.remove(p)
    p.destroy()

  def enter_step(self):
    D = self.core.DEBUG

    D('check breakpoints')

    for p in self.points:
      D(repr(p))

      if not p.active:
        D('inactive, not evaluating')
        continue

      if not p.is_triggered(self.core):
        D('not triggered, skipping')
        continue

      if p in self.triggered_points:
        D('already triggered by this step, ignore')
        continue

      if p.countdown > 0:
        p.countdown -= 1

      if p.countdown != 0:
        D('countdown %i, skip for now', p.countdown)
        continue

      self.core.INFO('Breakpoint triggered: %s', p)
      self.triggered_points.append(p)
      self.core.suspend()

      return True

    return False

  def exit_step(self):
    self.triggered_points = []

  def create_point(self, klass, *args, **kwargs):
    self.core.init_debug_set()

    p = klass(self, *args, **kwargs)
    self.add_point(p)

    return p

def cmd_bp_list(console, cmd):
  """
  List existing breakpoints
  """

  points = [
    ['ID', 'Active', 'Countdown', 'Core', 'Point']
  ]

  for point in itervalues(Point.points):
    points.append([
      point.id,
      '*' if point.active else '',
      point.countdown,
      point.debugging_set.core.cpuid_prefix,
      repr(point)
    ])

  console.table(points)

def cmd_bp_add_breakpoint(console, cmd):
  """
  Create new breakpoint: bp-break <#cpuid:#coreid> <address> [active] [countdown]
  """

  try:
    core = console.master.machine.core(cmd[1])

  except InvalidResourceError:
    console.write('go away')
    return

  ip = str2int(cmd[2])
  active = True if len(cmd) >= 3 and cmd[2] == 'yes' else False
  countdown = str2int(cmd[3]) if len(cmd) >= 4 else 0

  core.init_debug_set()
  point = core.debug.create_point(BreakPoint, ip, active = active, countdown = countdown)

  console.writeln('# OK: %s', point)

def cmd_bp_add_memory_watchpoint(console, cmd):
  """
  Create new memory watchpoint: bp-mwatch <#cpuid:#coreid> <address> [rw] [active] [countdown]'
  """

  try:
    core = console.master.machine.core(cmd[1])

  except InvalidResourceError:
    console.write('go away')
    return

  address = str2int(cmd[2])
  access = cmd[3] if len(cmd) >= 4 else 'r'
  active = True if len(cmd) >= 5 and cmd[4] == 'yes' else False
  countdown = str2int(cmd[5]) if len(cmd) >= 6 else 0

  core.init_debug_set()
  point = core.debug.create_point(MemoryWatchPoint, address, access, active = active, countdown = countdown)

  console.writeln('# OK: %s', point)

def cmd_bp_remove(console, cmd):
  """
  Remove breakpoint: bp-remove <id>
  """

  point = Point.points.get(int(cmd[1]))
  if point is None:
    console.writeln('go away')
    return

  point.debugging_set.remove_point(point)

  console.writeln('# OK')

def cmd_bp_active(console, cmd):
  """
  Toggle "active" flag for a breakpoint: bp-active <id>
  """

  point = Point.points.get(int(cmd[1]))
  if point is None:
    console.writeln('go away')
    return

  point.active = not point.active

  console.writeln('# OK: %s', point)

class VMDebugOperationList(enum.Enum):
  LOGGER_VERBOSITY = 0

class VMVerbosityLevels(enum.Enum):
  DEBUG = 0
  INFO = 1
  WARNING = 2
  ERROR = 3

VERBOSITY_LEVEL_MAP = {
  VMVerbosityLevels.DEBUG.value:   logging.DEBUG,
  VMVerbosityLevels.INFO.value:    logging.INFO,
  VMVerbosityLevels.WARNING.value: logging.WARNING,
  VMVerbosityLevels.ERROR.value:   logging.ERROR
}

class VMDebugInterrupt(IVirtualInterrupt):
  def run(self, core):
    from .cpu.registers import Registers

    core.DEBUG('VMDebugInterrupt: triggered')

    op = core.REG(Registers.R00).value
    core.REG(Registers.R00).value = 0xFFFF

    if op == VMDebugOperationList.LOGGER_VERBOSITY.value:
      verbosity = core.REG(Registers.R01).value

      if verbosity not in VERBOSITY_LEVEL_MAP:
        core.WARN('VMDebugInterrupt: unknown verbosity level: %s', verbosity)
        return

      core.LOGGER.setLevel(VERBOSITY_LEVEL_MAP[verbosity])
      core.DEBUG('VMDebugInterrupt: setting verbosity to %s', verbosity)

    else:
      core.WARN('VMDebugInterrupt: unknown operation requested: %s', op)

VIRTUAL_INTERRUPTS[IRQList.VMDEBUG.value] = VMDebugInterrupt



import sys
import optparse
import string

from six import itervalues

from ..snapshot import CoreDumpFile
from ..mm import ADDR_FMT, UINT16_FMT, UINT8_FMT, PAGE_SIZE, segment_addr_to_addr, addr_to_segment
from ..mm.binary import SectionFlags, File, SectionTypes
from ..cpu.registers import FlagsRegister, Registers
from ..log import WHITE, GREEN

def show_header(logger, state):
  state = state.get_child('machine')

  logger.info('=== Coredump header ===')
  logger.info('  # of CPUs:      %i', state.nr_cpus)
  logger.info('  # of CPU cores: %i', state.nr_cores)
  logger.info('  # of binaries:  %i', len(state.get_binary_states()))
  logger.info('')

def show_binaries(logger, state):
  logger.info('=== Binaries ===')

  for bs in state.get_child('machine').get_binary_states():
    logger.info('Path: %s', bs.path)
    logger.info('CS: %s DS: %s', UINT8_FMT(bs.cs), UINT8_FMT(bs.ds))
    logger.info('')

    table = [
      ['Section', 'Address', 'Size', 'Flags', 'First page', 'Last page']
    ]

    for rs in itervalues(bs.get_children()):
      flags = SectionFlags.from_uint16(rs.flags)

      table.append([rs.name, ADDR_FMT(rs.address), rs.size, flags.to_string(), rs.pages_start, rs.pages_start + rs.pages_cnt - 1])

    logger.table(table)
    logger.info('')

def show_cores(logger, state):
  logger.info('=== Cores ===')

  def __reg(index):
    return UINT16_FMT(cs.registers[index])

  for cs in state.get_child('machine').get_core_states():
    logger.info('Core #%i:#%i', cs.cpuid, cs.coreid)

    for i in range(0, Registers.REGISTER_SPECIAL, 4):
      regs = [(i + j) for j in range(0, 4) if (i + j) < Registers.REGISTER_SPECIAL]
      s = ['reg%02i=%s' % (reg, __reg(reg)) for reg in regs]
      logger.info('  %s', ' '.join(s))

    flags = FlagsRegister.from_uint16(cs.registers[Registers.FLAGS])

    logger.info('  cs=%s    ds=%s', __reg(Registers.CS), __reg(Registers.DS))
    logger.info('  fp=%s    sp=%s    ip=%s', __reg(Registers.FP), __reg(Registers.SP), __reg(Registers.IP))
    logger.info('  flags=%s', flags.to_string())
    logger.info('  cnt=%i, alive=%s, running=%s, idle=%s, exit=%i', cs.registers[Registers.CNT], cs.alive, cs.running, cs.idle, cs.exit_code)

    logger.info('')

def show_memory(logger, state):
  logger.info('=== Memory ===')

  state = state.get_child('machine').get_child('memory')

  logger.info('  Size:          %s', state.size)
  logger.info('  # of segments: %s', len(state.segments))
  logger.info('  # of pages:    %s', len(state.get_page_states()))
  logger.info('')

def show_stack(logger, state):
  logger.info('=== Stack ===')

  stacks = [pg for pg in state.get_child('machine').get_child('memory').get_page_states() if pg.stack == 1]

  sps = {}
  for cs in state.get_child('machine').get_core_states():
    sps[segment_addr_to_addr(cs.registers[Registers.DS], cs.registers[Registers.SP])] = '#%i:#%i' % (cs.cpuid, cs.coreid)

  for pg in stacks:
    pg_address = pg.index * PAGE_SIZE
    pg_segment = addr_to_segment(pg_address)

    logger.info('=== Page %s - %s %s ===', pg.index, UINT8_FMT(pg_segment), ADDR_FMT(pg_address))

    for ri in range(0, 16):
      row = []
      for ci in range(0, 8):
        offset = PAGE_SIZE - 2 - (ci * 16 + ri) * 2
        segment_offset = pg_address + offset

        v = pg.content[offset] | (pg.content[offset + 1] << 8)
        row.append((sps.get(segment_offset, None), v))

      logger.info('  ' + '  '.join(['%s %s' % (' ' if sp is None else '*', WHITE(UINT16_FMT(value)) if value != 0 else GREEN(UINT16_FMT(value))) for sp, value in row]))

    logger.info('')

def show_segments(logger, state):
  logger.info('=== Memory segments ===')

  for s in state.get_child('machine').get_child('memory').segments:
    logger.info('  Segment: %s', UINT8_FMT(s))

    for bs in state.get_child('machine').get_binary_states():
      if s == bs.cs:
        logger.info('    CS of %s', bs.path)
      if s == bs.ds:
        logger.info('    DS of %s', bs.path)

  logger.info('')

def show_pages(logger, state):
  logger.info('=== Memory pages ===')

  for pg in sorted(state.get_child('machine').get_child('memory').get_page_states(), key = lambda x: x.index):
    pg_addr = pg.index * PAGE_SIZE
    pg_segment = addr_to_segment(pg_addr)

    for bs in state.get_child('machine').get_binary_states():
      if pg_segment == bs.cs:
        break
      if pg_segment == bs.ds:
        break

    logger.info('  Page #%i (segment %s)', pg.index, UINT8_FMT(pg_segment))
    logger.info('    Flags: %s%s%s%s%s', 'r' if pg.read == 1 else '-', 'w' if pg.write == 1 else '-', 'x' if pg.execute == 1 else '-', 'd' if pg.dirty == 1 else '-', 'c' if pg.cache == 1 else '-')

    CPR = 32

    for i in range(0, 256 / CPR):
      s = []
      t = []

      for b in pg.content[CPR * i:CPR * (i + 1)]:
        c = '%02X' % b
        s.append(GREEN(c) if b == 0 else WHITE(c))

        c = chr(b)
        if c in string.printable[0:-5]:
          c = '%%' if c == '%' else (' ' + c)
          t.append(c)

        elif b == 10:
          t.append('\\n')

        elif b == 13:
          t.append('\\r')

        else:
          t.append('  ')

      logger.info('    ' + ADDR_FMT(pg_addr + i * CPR) + ': ' + ' '.join(s))
      logger.info('              ' + ' '.join(t))

      row_symbols = []
      for j in range(pg_addr + i * CPR, pg_addr + i * CPR + CPR):
        if j not in bs.symbols:
          continue

        row_symbols.append((j - (pg_addr + i * CPR)) * 3 * ' ' + '^' + ', '.join(bs.symbols[j]))

      for symbol in row_symbols:
        logger.info('              ' + symbol)

    logger.info('')

def load_binary_symbols(logger, vs, bs):
  bs.raw_binary = File.open(logger, bs.path, 'r')
  bs.raw_binary.load()

  bs.symbols = {}
  for i in range(0, bs.raw_binary.get_header().sections):
    s_header, s_content = bs.raw_binary.get_section(i)

    if s_header.type != SectionTypes.SYMBOLS:
      continue

    for entry in s_content:
      _header, _content = bs.raw_binary.get_section(entry.section)

      segment = bs.cs if _header.type == SectionTypes.TEXT else bs.ds

      symbol_name = bs.raw_binary.string_table.get_string(entry.name)
      symbol_address = segment_addr_to_addr(segment, entry.address)

      if symbol_address not in bs.symbols:
        bs.symbols[symbol_address] = []
      bs.symbols[symbol_address].append(symbol_name)

def main():
  from . import add_common_options, parse_options

  parser = optparse.OptionParser()
  add_common_options(parser)

  parser.add_option('-i', dest = 'file_in', default = None, help = 'Input file')

  parser.add_option('-H',         dest = 'header',   default = False, action = 'store_true', help = 'Show file header')
  parser.add_option('-C',         dest = 'cores',    default = False, action = 'store_true', help = 'Show cores')
  parser.add_option('-M',         dest = 'memory',   default = False, action = 'store_true', help = 'Show memory')
  parser.add_option('--segments', dest = 'segments', default = False, action = 'store_true', help = 'Show segments')
  parser.add_option('--pages',    dest = 'pages',    default = False, action = 'store_true', help = 'Show pages')
  parser.add_option('-b',         dest = 'binaries', default = False, action = 'store_true', help = 'Show binaries')
  parser.add_option('-s',         dest = 'stack',    default = False, action = 'store_true', help = 'Show stack content')
  parser.add_option('-a',         dest = 'all',      default = False, action = 'store_true', help = 'All of above')

  options, logger = parse_options(parser)

  if not options.file_in:
    parser.print_help()
    sys.exit(1)

  if options.all:
    options.header = options.cores = options.memory = options.segments = options.pages = options.binaries = options.stack = True

  logger.info('Input file: %s', options.file_in)

  with CoreDumpFile.open(logger, options.file_in, 'r') as f_in:
    state = f_in.load()

    for bs in state.get_child('machine').get_binary_states():
      load_binary_symbols(logger, state, bs)

    logger.info('')

    if options.header:
      show_header(logger, state)

    if options.binaries:
      show_binaries(logger, state)

    if options.cores:
      show_cores(logger, state)

    if options.memory:
      show_memory(logger, state)

    if options.segments:
      show_segments(logger, state)

    if options.stack:
      show_stack(logger, state)

    if options.pages:
      show_pages(logger, state)

if __name__ == '__main__':
  main()



import enum

from ctypes import c_ushort, c_uint
from ..util import Flags

class Registers(enum.IntEnum):
  # 16 16bit registers available
  R00   =  0
  R01   =  1
  R02   =  2
  R03   =  3
  R04   =  4
  R05   =  5
  R06   =  6
  R07   =  7
  R08   =  8
  R09   =  9
  R10   = 10
  R11   = 11
  R12   = 12

  # Some registers have special meaning and/or usage
  FP    = 13  # Frame Pointer
  SP    = 14  # Stack Pointer
  DS    = 15  # Data Segment Register
  CS    = 16  # Code Segment register
  IP    = 17  # Instruction pointer
  FLAGS = 18  # Flags
  CNT   = 19  # Instruction counter

  # First special register
  REGISTER_SPECIAL = 13

  # How many registers do we have? This many...
  REGISTER_COUNT = 20

PROTECTED_REGISTERS = [
  Registers.FP,
  Registers.DS,
  Registers.CS,
  Registers.IP,
  Registers.FLAGS,
  Registers.CNT
]

GENERAL_REGISTERS   = [r for r in Registers if r.value < Registers.REGISTER_SPECIAL.value]
RESETABLE_REGISTERS = [r for r in Registers if r not in (Registers.FLAGS, Registers.REGISTER_SPECIAL, Registers.REGISTER_COUNT)]

REGISTER_NAMES = ['r{}'.format(r.value) for r in Registers if r.value < Registers.REGISTER_SPECIAL.value] + ['fp', 'sp', 'ds', 'cs', 'ip', 'flags', 'cnt']

class FlagsRegister(Flags):
  _fields_ = [
    ('privileged', c_ushort, 1),
    ('hwint',      c_ushort, 1),
    ('e',          c_ushort, 1),
    ('z',          c_ushort, 1),
    ('o',          c_ushort, 1),
    ('s',          c_ushort, 1)
  ]

  flag_labels = 'PHEZOS'

class RegisterSet(object):
  def __init__(self):
    super(RegisterSet, self).__init__()

    self.map = {}

    for register_name, register_id in zip(REGISTER_NAMES, Registers):
      if register_name == 'cnt':
        register_class = c_uint

      elif register_name == 'flags':
        register_class = FlagsRegister

      else:
        register_class = c_ushort

      register = register_class()

      setattr(self, register_name, register)
      self.map[register_name] = register
      self.map[register_id.value] = register



"""
Stack-based coprocessor, providing several arithmetic operations with "long",
32 bits wide integers.

Coprocessor's instructions operates above stack of 8 long ints. Operations to
move values between math stack and regular registers are also available.

Operations usually consume their operands from the stack, and push the result
back.
"""

import enum

from ...interfaces import ISnapshotable, IVirtualInterrupt
from . import Coprocessor
from .. import CPUException
from ..registers import Registers
from ..instructions import InstructionSet, InstDescriptor_Generic, Inst_POP, Inst_PUSH, Inst_MOV, Inst_SIS, INSTRUCTION_SETS
from ...mm import u32, i32, UINT32_FMT, UINT16_FMT
from ...devices import IRQList, VIRTUAL_INTERRUPTS
from ...snapshot import SnapshotNode

#: Number of available spots on the math stack.
STACK_DEPTH = 8

class MathCoprocessorState(SnapshotNode):
  """
  Snapshot node holding the state of math coprocessor.
  """

  def __init__(self):
    super(MathCoprocessorState, self).__init__('stack')

class MathOperationList(enum.IntEnum):
  """
  List of available arithmetic operations.
  """

  INCL  =  0
  DECL  =  1
  ADDL  =  2
  SUBL  =  3
  MULL  =  4
  DIVL  =  5
  MODL  =  6
  ITOL  =  7
  LTOI  =  8
  LTOII =  9
  IITOL = 10
  DUPL  = 11
  UTOL  = 12
  SYMDIVL = 13
  SYMMODL = 14

class EmptyMathStackError(CPUException):
  """
  Raised when operation expects at least one value on math stack but stack
  is empty.
  """

  def __init__(self, *args, **kwargs):
    super(EmptyMathStackError, self).__init__('Math stack is empty', *args, **kwargs)

class FullMathStackError(CPUException):
  """
  Raised when operation tries to put value on math stack but there is no empty
  spot available.
  """

  def __init__(self, *args, **kwargs):
    super(FullMathStackError, self).__init__('Math stack is full', *args, **kwargs)

class RegisterSet(ISnapshotable, object):
  """
  Math stack wrapping class. Provides basic push/pop access, and direct access
  to a top of the stack.

  :param ducky.cpu.CPUCore core: CPU core registers belong to.
  """

  def __init__(self, core):
    super(RegisterSet, self).__init__()

    self.core = core
    self.stack = []

  def save_state(self, parent):
    self.core.DEBUG('RegisterSet.save_state')

    state = parent.add_child('math_coprocessor', MathCoprocessorState())

    state.stack = []
    for lr in self.stack:
      state.stack.append(int(lr.value))

  def load_state(self, state):
    self.core.DEBUG('RegisterSet.load_state')

    for lr in state.depth:
      self.stack.append(u32(lr))

  def push(self, v):
    """
    Push new value on top of the stack.

    :raises ducky.cpu.coprocessor.math_copro.FullMathStackError: if there is no space available on the stack.
    """

    if len(self.stack) == STACK_DEPTH:
      raise FullMathStackError()

    self.stack.append(v)

  def pop(self):
    """
    Pop the top value from stack and return it.

    :raises ducky.cpu.coprocessor.math_copro.EmptyMathStackError: if there are no values on the stack.
    """

    if not self.stack:
      raise EmptyMathStackError()

    return self.stack.pop()

  def tos(self):
    """
    Return the top of the stack, without removing it from a stack.

    :raises ducky.cpu.coprocessor.math_copro.EmptyMathStackError: if there are no values on the stack.
    """

    if not self.stack:
      raise EmptyMathStackError()

    return self.stack[-1]

class MathCoprocessor(ISnapshotable, Coprocessor):
  """
  Coprocessor itself, includes its register set ("math stack").

  :param ducky.cpu.CPUCore core: CPU core coprocessor belongs to.
  """

  def __init__(self, core, *args, **kwargs):
    super(MathCoprocessor, self).__init__(core, *args, **kwargs)

    self.registers = RegisterSet(core)

  def save_state(self, parent):
    self.registers.save_state(parent)

  def load_state(self, state):
    self.registers.load_state(state)

  def dump_stack(self):
    """
    Log content of the stack using parent's ``DEBUG`` method.
    """

    D = self.core.DEBUG

    D('Math stack:')

    for index, lr in enumerate(self.registers.stack):
      D('#%02i: %s', index, UINT32_FMT(lr.value))

    D('---')

  def op_itol(self):
    """
    Extend 16bit int to 32bit long int. Int is read from `r1`. Signed extension
    is used.
    """

    D = self.core.DEBUG

    r = self.core.REG(Registers.R01)
    lr = u32(r.value)

    if r.value & 0x8000 != 0:
      lr.value |= 0xFFFF0000

    self.registers.push(lr)

    D('itol: i=%s', UINT16_FMT(r.value))
    self.dump_stack()

  def op_utol(self):
    """
    Extend 16bit int to 32bit long int. Int is read from `r1`.
    """

    D = self.core.DEBUG

    r = self.core.REG(Registers.R01)
    lr = u32(r.value)

    self.registers.push(lr)

    D('utol: i=%s', UINT16_FMT(r.value))
    self.dump_stack()

  def op_iitol(self):
    """
    Merge two 16bit ints to 32bit long int. Ints are read from `r1` (lower
    bytes) and `r2` (higher bytes).
    """

    D = self.core.DEBUG

    r1 = self.core.REG(Registers.R01)
    r2 = self.core.REG(Registers.R02)
    lr = u32(r1.value | (r2.value << 16))

    self.registers.push(lr)

    D('iitol: lb=%s, hb=%s', UINT16_FMT(r1.value), UINT16_FMT(r2.value))
    self.dump_stack()

  def op_ltoi(self):
    """
    Truncate 32bit long int to 16bit int. New int is stored in `r1` register.
    """

    D = self.core.DEBUG

    lr = self.registers.pop()

    self.core.check_protected_reg(Registers.R01)
    self.core.REG(Registers.R01).value = lr.value & 0xFFFF

    D('ltoi: i=%s', UINT16_FMT(self.core.REG(Registers.R01).value))
    self.dump_stack()

  def op_ltoii(self):
    """
    Split 32bit long into to two 16bit ints. New ints are stored in `r1` (lower
    bytes) and `r2` (higher bytes).
    """

    D = self.core.DEBUG

    r1 = self.core.REG(Registers.R01)
    r2 = self.core.REG(Registers.R02)
    lr = self.registers.pop()

    self.core.check_protected_reg(Registers.R01)
    self.core.check_protected_reg(Registers.R02)
    r1.value =  lr.value & 0x0000FFFF
    r2.value = (lr.value & 0xFFFF0000) >> 16

    D('ltoii: lb=%s, hb=%s', UINT16_FMT(r1.value), UINT16_FMT(r2.value))
    self.dump_stack()

  def op_dupl(self):
    """
    Duplicate top of the math stack.
    """

    D = self.core.DEBUG

    lr = self.registers.tos()

    self.registers.push(u32(lr.value))

    D('dupl:')
    self.dump_stack()

  def op_incl(self):
    """
    Increment top of the stack by one.
    """

    self.registers.tos().value += 1

  def op_decl(self):
    """
    Decrement top of the stack by one.
    """

    self.registers.tos().value -= 1

  def op_addl(self):
    """
    Add the topmost value to the value below.
    """

    lr = self.registers.pop()

    self.registers.tos().value += lr.value

  def op_subl(self):
    """
    Subtract the top value from the value bellow.
    """

    lr = self.registers.pop()

    self.registers.tos().value -= lr.value

  def op_mull(self):
    """
    Multiply two topmost values.
    """

    D = self.core.DEBUG

    lr = self.registers.pop()
    old_tos = self.registers.tos().value

    self.registers.tos().value *= lr.value

    D('mull: %s %s', UINT32_FMT(old_tos), UINT32_FMT(lr.value))
    self.dump_stack()

  def op_divl(self):
    """
    Divide the value below the top of the math stack by the topmost value.
    """

    D = self.core.DEBUG

    D('divl:')

    lr = self.registers.pop()
    old_tos = self.registers.tos().value
    tos = self.registers.tos()

    i = i32(tos.value).value
    j = i32(lr.value).value
    D('  i=%i, j=%i (%s, %s)', i, j, type(i), type(j))
    i //= j
    D('  i=%i (%s)', i, type(i))
    tos.value = i

    D('divl: %s %s', UINT32_FMT(old_tos), UINT32_FMT(lr.value))
    self.dump_stack()

  def op_symdivl(self):
    """
    The same operation like ``DIVL`` but provides symmetric results.
    """

    D = self.core.DEBUG

    D('symdivl:')

    lr = self.registers.pop()
    old_tos = self.registers.tos().value
    tos = self.registers.tos()

    i = i32(tos.value).value
    j = i32(lr.value).value
    D('  i=%i, j=%i (%s, %s)', i, j, type(i), type(j))
    i = int(float(i) / float(j))
    D('  i=%i (%s)', i, type(i))
    tos.value = i

    D('divl: %s %s', UINT32_FMT(old_tos), UINT32_FMT(lr.value))
    self.dump_stack()

  def op_modl(self):
    """
    Computes remainder after division. The value below the top of the math
    stack is divided by the topmost value, and remainder is pushed back to
    stack.
    """

    D = self.core.DEBUG

    D('modl:')

    lr = self.registers.pop()
    old_tos = self.registers.tos().value
    tos = self.registers.tos()

    i = i32(tos.value).value
    j = i32(lr.value).value
    D('  i=%i, j=%i (%s, %s)', i, j, type(i), type(j))
    i %= j
    D('  i=%i (%s)', i, type(i))
    tos.value = i

    D('modl: %s %s', UINT32_FMT(old_tos), UINT32_FMT(lr.value))
    self.dump_stack()

  def op_symmodl(self):
    """
    The same operations as ``MODL`` but provides symmetric results.
    """

    import math

    D = self.core.DEBUG

    D('symmodl:')

    lr = self.registers.pop()
    old_tos = self.registers.tos().value
    tos = self.registers.tos()

    i = i32(tos.value).value
    j = i32(lr.value).value
    D('  i=%i, j=%i (%s, %s)', i, j, type(i), type(j))
    i = int(math.fmod(i, j))
    D('  i=%i (%s)', i, type(i))
    tos.value = i

    D('modl: %s %s', UINT32_FMT(old_tos), UINT32_FMT(lr.value))
    self.dump_stack()

#
# Virtual interrupt
#
class MathInterrupt(IVirtualInterrupt):
  """
  Virtual interrupt handler of math coprocessor.
  """

  def run(self, core):
    """
    Called when coprocessor's interrupt is triggered. Dispatches call to a proper
    method to handle requested operation.

    Id of requested operation is expected in `r0`.
    """

    core.DEBUG('MathInterrupt: triggered')

    op = core.REG(Registers.R00).value
    core.REG(Registers.R00).value = 0

    if op == MathOperationList.ITOL:
      core.math_coprocessor.op_itol()

    elif op == MathOperationList.UTOL:
      core.math_coprocessor.op_utol()

    elif op == MathOperationList.IITOL:
      core.math_coprocessor.op_iitol()

    elif op == MathOperationList.LTOI:
      core.math_coprocessor.op_ltoi()

    elif op == MathOperationList.LTOII:
      core.math_coprocessor.op_ltoii()

    elif op == MathOperationList.DUPL:
      core.math_coprocessor.op_dupl()

    elif op == MathOperationList.INCL:
      core.math_coprocessor.op_incl()

    elif op == MathOperationList.DECL:
      core.math_coprocessor.op_decl()

    elif op == MathOperationList.ADDL:
      core.math_coprocessor.op_addl()

    elif op == MathOperationList.SUBL:
      core.math_coprocessor.op_subl()

    elif op == MathOperationList.MULL:
      core.math_coprocessor.op_mull()

    elif op == MathOperationList.DIVL:
      core.math_coprocessor.op_divl()

    elif op == MathOperationList.MODL:
      core.math_coprocessor.op_modl()

    elif op == MathOperationList.SYMDIVL:
      core.math_coprocessor.op_symdivl()

    elif op == MathOperationList.SYMMODL:
      core.math_coprocessor.op_symmodl()

    else:
      core.WARN('Unknown math operation requested: %s', op)
      core.REG(Registers.R00).value = 0xFFFF

#
# Instruction set
#
class MathCoprocessorOpcodes(enum.IntEnum):
  """
  Math coprocessor's instruction opcodes.
  """

  INCL    =  0
  DECL    =  1
  ADDL    =  2
  SUBL    =  3
  MULL    =  4
  DIVL    =  5
  MODL    =  6
  ITOL    =  7
  LTOI    =  8
  LTOII   =  9
  IITOL   = 10
  DUPL    = 11
  UTOL    = 12
  SYMDIVL = 13
  SYMMODL = 14

  POP     = 60
  PUSH    = 61
  MOV     = 62
  SIS     = 63

class MathCoprocessorInstructionSet(InstructionSet):
  """
  Math coprocessor's instruction set.
  """

  instruction_set_id = 1

  opcodes = MathCoprocessorOpcodes

class Inst_Math_PUSH(Inst_PUSH):
  opcode = MathCoprocessorOpcodes.PUSH

class Inst_Math_POP(Inst_POP):
  opcode = MathCoprocessorOpcodes.POP

class Inst_Math_MOV(Inst_MOV):
  opcode = MathCoprocessorOpcodes.MOV

class Inst_INCL(InstDescriptor_Generic):
  mnemonic = 'incl'
  opcode = MathCoprocessorOpcodes.INCL

  @staticmethod
  def execute(core, inst):
    core.math_coprocessor.op_incl()

class Inst_DECL(InstDescriptor_Generic):
  mnemonic = 'decl'
  opcode = MathCoprocessorOpcodes.DECL

  @staticmethod
  def execute(core, inst):
    core.math_coprocessor.op_decl()

class Inst_ADDL(InstDescriptor_Generic):
  mnemonic = 'addl'
  opcode = MathCoprocessorOpcodes.ADDL

  @staticmethod
  def execute(core, inst):
    core.math_coprocessor.op_addl()

class Inst_SUBL(InstDescriptor_Generic):
  mnemonic = 'subl'
  opcode = MathCoprocessorOpcodes.SUBL

  @staticmethod
  def execute(core, inst):
    core.math_coprocessor.op_subl()

class Inst_MULL(InstDescriptor_Generic):
  mnemonic = 'mull'
  opcode = MathCoprocessorOpcodes.MULL

  @staticmethod
  def execute(core, inst):
    core.math_coprocessor.op_mull()

class Inst_DIVL(InstDescriptor_Generic):
  mnemonic = 'divl'
  opcode = MathCoprocessorOpcodes.DIVL

  @staticmethod
  def execute(core, inst):
    core.math_coprocessor.op_divl()

class Inst_MODL(InstDescriptor_Generic):
  mnemonic = 'modl'
  opcode = MathCoprocessorOpcodes.MODL

  @staticmethod
  def execute(core, inst):
    core.math_coprocessor.op_modl()

class Inst_ITOL(InstDescriptor_Generic):
  mnemonic = 'itol'
  opcode = MathCoprocessorOpcodes.ITOL

  @staticmethod
  def execute(core, inst):
    core.math_coprocessor.op_itol()

class Inst_LTOI(InstDescriptor_Generic):
  mnemonic = 'ltoi'
  opcode = MathCoprocessorOpcodes.LTOI

  @staticmethod
  def execute(core, inst):
    core.math_coprocessor.op_ltoi()

class Inst_LTOII(InstDescriptor_Generic):
  mnemonic = 'ltoii'
  opcode = MathCoprocessorOpcodes.LTOII

  @staticmethod
  def execute(core, inst):
    core.math_coprocessor.op_ltoii()

class Inst_IITOL(InstDescriptor_Generic):
  mnemonic = 'iitol'
  opcode = MathCoprocessorOpcodes.IITOL

  @staticmethod
  def execute(core, inst):
    core.math_coprocessor.op_iitol()

class Inst_DUPL(InstDescriptor_Generic):
  mnemonic = 'dupl'
  opcode = MathCoprocessorOpcodes.DUPL

  @staticmethod
  def execute(core, inst):
    core.math_coprocessor.op_dupl()

class Inst_UTOL(InstDescriptor_Generic):
  mnemonic = 'utol'
  opcode = MathCoprocessorOpcodes.UTOL

  @staticmethod
  def execute(core, inst):
    core.math_coprocessor.op_utol()

class Inst_SYMDIVL(InstDescriptor_Generic):
  mnemonic = 'symdivl'
  opcode = MathCoprocessorOpcodes.SYMDIVL

  @staticmethod
  def execute(core, inst):
    core.math_coprocessor.op_symdivl()

class Inst_SYMMODL(InstDescriptor_Generic):
  mnemonic = 'symmodl'
  opcode = MathCoprocessorOpcodes.SYMMODL

  @staticmethod
  def execute(core, inst):
    core.math_coprocessor.op_symmodl()

Inst_INCL(MathCoprocessorInstructionSet)
Inst_DECL(MathCoprocessorInstructionSet)
Inst_ADDL(MathCoprocessorInstructionSet)
Inst_SUBL(MathCoprocessorInstructionSet)
Inst_MULL(MathCoprocessorInstructionSet)
Inst_DIVL(MathCoprocessorInstructionSet)
Inst_MODL(MathCoprocessorInstructionSet)
Inst_ITOL(MathCoprocessorInstructionSet)
Inst_LTOI(MathCoprocessorInstructionSet)
Inst_LTOII(MathCoprocessorInstructionSet)
Inst_IITOL(MathCoprocessorInstructionSet)
Inst_DUPL(MathCoprocessorInstructionSet)
Inst_UTOL(MathCoprocessorInstructionSet)
Inst_SYMDIVL(MathCoprocessorInstructionSet)
Inst_SYMMODL(MathCoprocessorInstructionSet)
Inst_Math_PUSH(MathCoprocessorInstructionSet)
Inst_Math_POP(MathCoprocessorInstructionSet)
Inst_Math_MOV(MathCoprocessorInstructionSet)
Inst_SIS(MathCoprocessorInstructionSet)

MathCoprocessorInstructionSet.init()

INSTRUCTION_SETS[MathCoprocessorInstructionSet.instruction_set_id] = MathCoprocessorInstructionSet
VIRTUAL_INTERRUPTS[IRQList.MATH.value] = MathInterrupt



import collections
import ctypes
import functools
import mmap
import os.path
import re

from six import iteritems, itervalues, integer_types, string_types, PY2

from .. import cpu
from .. import mm
from ..cpu.coprocessor.math_copro import MathCoprocessorInstructionSet  # noqa - it's not unused, SIS instruction may need it but that's hidden from flake

from ..mm import UInt8, UInt16, ADDR_FMT, PAGE_SIZE
from ..mm.binary import SectionTypes, SectionFlags, SymbolFlags, RelocFlags
from ..util import align, str2bytes

align_to_next_page = functools.partial(align, PAGE_SIZE)
align_to_next_mmap = functools.partial(align, mmap.PAGESIZE)

RE_COMMENT = re.compile(r'^\s*[/;].*?$', re.MULTILINE)
RE_INCLUDE = re.compile(r'^\s*\.include\s+"(?P<file>[a-zA-Z0-9_\-/\.]+)\s*"$', re.MULTILINE)
RE_IFDEF = re.compile(r'^\s*\.ifdef\s+(?P<var>[a-zA-Z0-9_]+)\s*$', re.MULTILINE)
RE_IFNDEF = re.compile(r'^\s*\.ifndef\s+(?P<var>[a-zA-Z0-9_]+)\s*$', re.MULTILINE)
RE_ELSE = re.compile(r'^\s*\.else\s*$', re.MULTILINE)
RE_ENDIF = re.compile(r'^\s*\.endif\s*$', re.MULTILINE)
RE_VAR_DEF = re.compile(r'^\s*\.def\s+(?P<var_name>[a-zA-Z][a-zA-Z0-9_]*):\s*(?P<var_body>.*?)$', re.MULTILINE)
RE_MACRO_DEF = re.compile(r'^\s*\.macro\s+(?P<macro_name>[a-zA-Z][a-zA-Z0-9_]*)(?:\s+(?P<macro_params>.*?))?:$', re.MULTILINE | re.DOTALL)
RE_MACRO_END = re.compile(r'^\s*\.end\s*$', re.MULTILINE)
RE_ASCII = re.compile(r'^\s*\.ascii\s+"(?P<value>.*?)"\s*$', re.MULTILINE)
RE_BYTE = re.compile(r'^\s*\.byte\s+(?:(?P<value_hex>-?0x[a-fA-F0-9]+)|(?P<value_dec>(?:0)|(?:-?[1-9][0-9]*))|(?P<value_var>[a-zA-Z][a-zA-Z0-9_]*))\s*$', re.MULTILINE)
RE_DATA = re.compile(r'^\s*\.data(?:\s+(?P<name>\.[a-z][a-z0-9_]*))?\s*$', re.MULTILINE)
RE_INT = re.compile(r'^\s*\.int\s+(?:(?P<value_hex>-?0x[a-fA-F0-9]+)|(?P<value_dec>0|(?:-?[1-9][0-9]*))|(?P<value_var>[a-zA-Z][a-zA-Z0-9_]*)|(?P<value_label>&[a-zA-Z_\.][a-zA-Z0-9_]*))\s*$', re.MULTILINE)
RE_SECTION = re.compile(r'^\s*\.section\s+(?P<name>\.[a-zA-z0-9_]+)(?:,\s*(?P<flags>[rwxlbmg]*))?\s*$', re.MULTILINE)
RE_SET = re.compile(r'^\s*\.set\s+(?P<name>[a-zA-Z_][a-zA-Z0-9_]*),\s*(?:(?P<current>\.)|(?P<value_hex>-?0x[a-fA-F0-9]+)|(?P<value_dec>0|(?:-?[1-9][0-9]*))|(?P<value_label>&[a-zA-Z][a-zA-Z0-9_]*))\s*$', re.MULTILINE)
RE_SIZE = re.compile(r'^\s*\.size\s+(?P<size>[1-9][0-9]*)\s*$', re.MULTILINE)
RE_SPACE = re.compile(r'^\s*\.space\s+(?P<size>[1-9][0-9]*)\s*$', re.MULTILINE)
RE_STRING = re.compile(r'^\s*\.string\s+"(?P<value>.*?)"\s*$', re.MULTILINE)
RE_TEXT = re.compile(r'^\s*\.text(?:\s+(?P<name>\.[a-z][a-z0-9_]*))?\s*$', re.MULTILINE)
RE_TYPE = re.compile(r'^\s*\.type\s+(?P<name>[a-zA-Z_\.][a-zA-Z0-9_]*),\s*(?P<type>(?:char|byte|int|ascii|string|space))\s*$', re.MULTILINE)
RE_GLOBAL = re.compile(r'^\s*.global\s+(?P<name>[a-zA-Z_][a-zA-Z0-9_]*)\s*$', re.MULTILINE)

class AssemblerError(Exception):
  def __init__(self, filename, lineno, msg, line):
    super(AssemblerError, self).__init__('{}:{}: {}'.format(filename, lineno, msg))

    self.filename = filename
    self.lineno   = lineno
    self.msg      = msg
    self.line     = line

class IncompleteDirectiveError(AssemblerError):
  def __init__(self, filename, lineno, msg, line):
    super(IncompleteDirectiveError, self).__init__(filename, lineno, 'Incomplete directive: %s' % msg, line)

class UnknownFileError(AssemblerError):
  def __init__(self, filename, lineno, msg, line):
    super(UnknownFileError, self).__init__(filename, lineno, 'Unknown file: %s' % msg, line)

class Buffer(object):
  def __init__(self, logger, filename, buff):
    super(Buffer, self).__init__()

    self.logger = logger
    self.DEBUG = logger.debug
    self.INFO = logger.info
    self.WARN = logger.warning
    self.ERROR = logger.error
    self.EXCEPTION = logger.exception

    self.buff = buff

    self.filename = filename
    self.lineno = 0
    self.last_line = None

  def get_line(self):
    while self.buff:
      self.lineno += 1

      line = self.buff.pop(0)

      if isinstance(line, tuple):
        self.lineno = line[1]
        self.filename = line[0]

        self.DEBUG('buffer: file switch: filename=%s, lineno=%s', self.filename, self.lineno)
        continue

      if not line:
        continue

      self.DEBUG('buffer: new line %s:%s: %s', self.filename, self.lineno, line)
      self.last_line = line
      return line

    self.last_line = None
    return None

  def put_line(self, line):
    self.buff.insert(0, line)
    self.lineno -= 1

  def put_buffer(self, buff, filename = None):
    filename = filename or '<unknown>'

    self.buff.insert(0, (self.filename, self.lineno))

    if isinstance(buff, string_types):
      buff = buff.split('\n')

    for line in reversed(buff):
      self.buff.insert(0, line)

    self.buff.insert(0, (filename, 0))

  def has_lines(self):
    return len(self.buff) > 0

  def get_error(self, cls, msg):
    return cls(self.filename, self.lineno, msg, self.last_line)

class Section(object):
  def __init__(self, s_name, s_type, s_flags):
    super(Section, self).__init__()

    self.name    = s_name
    self.type    = s_type
    self.flags   = s_flags
    self.content = []

    self.base = None
    self.ptr  = UInt16(0)

  def __getattr__(self, name):
    if name == 'data_size':
      return sum([sizeof(i) for i in self.content])

    if name == 'file_size':
      return align_to_next_mmap(self.data_size) if self.flags.mmapable == 1 else self.data_size

    if name == 'items':
      return len(self.content)

  def __repr__(self):
    return '<Section: name={}, type={}, flags={}, base={}, ptr={}, items={}, data_size={}, file_size={}>'.format(self.name, self.type, self.flags.to_string(), self.base, self.ptr, self.items, self.data_size, self.file_size)

class TextSection(Section):
  def __init__(self, s_name, flags = None, **kwargs):
    super(TextSection, self).__init__(s_name, SectionTypes.TEXT, flags or SectionFlags.create(readable = True, executable = True, loadable = True))

class RODataSection(Section):
  def __init__(self, s_name, flags = None, **kwargs):
    super(RODataSection, self).__init__(s_name, SectionTypes.DATA, flags or SectionFlags.create(readable = True, loadable = True))

class DataSection(Section):
  def __init__(self, s_name, flags = None, **kwargs):
    super(DataSection, self).__init__(s_name, SectionTypes.DATA, flags or SectionFlags.create(readable = True, writable = True, loadable = True))

class BssSection(Section):
  def __init__(self, s_name, flags = None, **kwargs):
    super(BssSection, self).__init__(s_name, SectionTypes.DATA, flags or SectionFlags.create(readable = True, writable = True, loadable = True, bss = True))

class SymbolsSection(Section):
  def __init__(self, s_name, flags = None, **kwargs):
    super(SymbolsSection, self).__init__(s_name, SectionTypes.SYMBOLS, SectionFlags.create())

class RelocSection(Section):
  def __init__(self, s_name, flags = None, **kwargs):
    super(RelocSection, self).__init__(s_name, SectionTypes.RELOC, SectionFlags.create())

class Label(object):
  def __init__(self, name, section, filename, lineno):
    super(Label, self).__init__()

    self.name = name
    self.section = section

    self.filename = filename
    self.lineno = lineno

  def __repr__(self):
    return '<label {} in section {} ({}:{})>'.format(self.name, self.section.name if self.section else None, self.filename, self.lineno)

class RelocSlot(object):
  def __init__(self, name, flags = None, patch_section = None, patch_address = None, patch_offset = None, patch_size = None):
    super(RelocSlot, self).__init__()

    self.name = name
    self.flags = flags or RelocFlags()
    self.patch_section = patch_section
    self.patch_address = patch_address
    self.patch_offset = patch_offset
    self.patch_size = patch_size

    self.size = UInt16(0)

  def __repr__(self):
    return '<RelocSlot: name=%s, flags=%s, section=%s, address=%s, offset=%s, size=%s>' % (self.name, self.flags.to_string(), self.patch_section, ADDR_FMT(self.patch_address), self.patch_offset, self.patch_size)

class DataSlot(object):
  def __init__(self):
    super(DataSlot, self).__init__()

    self.name  = None
    self.size  = None
    self.refers_to = None
    self.value = None

    self.flags = SymbolFlags()

    self.section = None
    self.section_ptr = None

    self.filename = None
    self.lineno = None

  def close(self):
    pass

class ByteSlot(DataSlot):
  symbol_type = mm.binary.SymbolDataTypes.CHAR

  def close(self):
    self.size = UInt16(1)

    if self.refers_to:
      return

    self.value = UInt8(self.value or 0)

  def __repr__(self):
    return '<ByteSlot: name={}, size={}, section={}, value={}>'.format(self.name, self.size, self.section.name if self.section else '', self.value)

class IntSlot(DataSlot):
  symbol_type = mm.binary.SymbolDataTypes.INT

  def close(self):
    self.size = UInt16(2)

    if self.refers_to:
      return

    self.value = UInt16(self.value or 0)
    self.size = UInt16(2)

  def __repr__(self):
    return '<IntSlot: name={}, size={}, section={}, value={}, refers_to={}>'.format(self.name, self.size, self.section.name if self.section else '', self.value, self.refers_to)

class CharSlot(DataSlot):
  symbol_type = mm.binary.SymbolDataTypes.CHAR

  def close(self):
    self.value = UInt8(ord(self.value or '\0'))
    self.size = UInt16(1)

  def __repr__(self):
    return '<CharSlot: name={}, section={}, value={}>'.format(self.name, self.section.name if self.section else '', self.value)

class SpaceSlot(DataSlot):
  symbol_type = mm.binary.SymbolDataTypes.ASCII

  def close(self):
    self.value = None
    self.size = UInt16(self.size)

  def __repr__(self):
    return '<SpaceSlot: name={}, size={}, section={}>'.format(self.name, self.size, self.section.name if self.section else '')

class AsciiSlot(DataSlot):
  symbol_type = mm.binary.SymbolDataTypes.ASCII

  def close(self):
    self.value = self.value or ''
    self.value = [UInt8(ord(c)) for c in self.value]
    self.size = UInt16(len(self.value))

  def __repr__(self):
    return '<AsciiSlot: name={}, size={}, section={}, value={}>'.format(self.name, self.size, self.section.name if self.section else '', self.value)

class StringSlot(DataSlot):
  symbol_type = mm.binary.SymbolDataTypes.STRING

  def close(self):
    self.value = self.value or ''
    self.value = [UInt8(ord(c)) for c in self.value] + [UInt8(0)]
    self.size = UInt16(len(self.value))

  def __repr__(self):
    return '<StringSlot: name={}, size={}, section={}, value={}>'.format(self.name, self.size, self.section.name if self.section else '', self.value)

class FunctionSlot(DataSlot):
  symbol_type = mm.binary.SymbolDataTypes.FUNCTION

  def close(self):
    self.size = UInt16(0)

  def __repr__(self):
    return '<FunctionSlot: name={}, section={}>'.format(self.name, self.section.name if self.section else '')

def sizeof(o):
  if isinstance(o, RelocSlot):
    return 0

  if isinstance(o, DataSlot):
    return o.size.u16

  if isinstance(o, ctypes.LittleEndianStructure):
    return ctypes.sizeof(o)

  return None

if PY2:
  def decode_string(s):
    return s.decode('string_escape')

else:
  def decode_string(s):
    return str2bytes(s).decode('unicode_escape')

def translate_buffer(logger, buff, base_address = None, mmapable_sections = False, writable_sections = False, filename = None, defines = None, includes = None):
  DEBUG = logger.debug

  filename = filename or '<unknown>'
  defines = defines or []
  includes = includes or []
  includes.insert(0, os.getcwd())

  buff = Buffer(logger, filename, buff.split('\n'))

  base_address = base_address or UInt16(0)

  sections_pass1 = collections.OrderedDict([
    ('.text',   TextSection('.text')),
    ('.rodata', RODataSection('.rodata')),
    ('.data',   DataSection('.data')),
    ('.bss',    BssSection('.bss')),
    ('.symtab', SymbolsSection('.symtab')),
    ('.reloc',  RelocSection('.reloc'))
  ])

  if mmapable_sections:
    for section in itervalues(sections_pass1):
      section.flags.mmapable = 1

  if writable_sections:
    for section in [_section for _section in itervalues(sections_pass1) if _section.name in ('.text', '.rodata', '.data', '.bss')]:
      section.flags.writable = 1

  DEBUG('Pass #1')

  labeled = []

  line = None
  lineno = None

  def __apply_defs(line):
    orig_line = line

    for def_pattern, def_value in iteritems(defs):
      line = def_pattern.sub(def_value, line)

    if orig_line != line:
      DEBUG(msg_prefix + 'variables replaced: line="%s"', line)

    return line

  def __apply_macros(line):
    for m_pattern, m_desc in iteritems(macros):
      matches = m_pattern.match(line)
      if not matches:
        continue

      DEBUG(msg_prefix + 'replacing macro: name=%s', m_desc['name'])

      if m_desc['params']:
        matches = matches.groupdict()

        replace_map = {}
        for i in range(0, len(m_desc['params'])):
          replace_map[re.compile(r'#{}'.format(m_desc['params'][i]))] = matches['arg{}'.format(i)]

        DEBUG(msg_prefix + 'macro args: %s', ', '.join(['{} => {}'.format(pattern.pattern, repl) for pattern, repl in iteritems(replace_map)]))

        body = []
        for line in m_desc['body']:
          for pattern, repl in iteritems(replace_map):
            line = pattern.sub(repl, line)
          body.append(line)

        buff.put_buffer(body)

      else:
        buff.put_buffer(m_desc['body'])

      return True

    return False

  def __get_refers_to_operand(inst):
    r_address = references[inst.refers_to].section_ptr.u16

    if inst.refers_to.startswith(''):
      r_address -= (inst.address.u16 + 4)

    return r_address

  def __parse_byte(var, matches):
    if not var.lineno:
      var.filename = buff.filename
      var.lineno = buff.lineno

    matches = matches.groupdict()

    v_value = matches.get('value_dec')
    if v_value:
      var.value = int(v_value)
      return

    v_value = matches.get('value_hex')
    if v_value:
      var.value = int(v_value, base = 16)
      return

    v_value = matches.get('value_var')
    if v_value:
      referred_var = variables[matches['value_var']]

      if isinstance(referred_var, integer_types):
        var.value = referred_var
      else:
        var.refers_to = referred_var

      return

    raise buff.get_error(IncompleteDirectiveError, '.byte directive without a meaningful value')

  def __parse_int(var, matches):
    if not var.lineno:
      var.filename = buff.filename
      var.lineno = buff.lineno

    matches = matches.groupdict()

    v_value = matches.get('value_dec')
    if v_value:
      var.value = int(v_value)
      return

    v_value = matches.get('value_hex')
    if v_value:
      var.value = int(v_value, base = 16)
      return

    v_value = matches.get('value_var')
    if v_value:
      if matches['value_var'] not in variables:
        raise buff.get_error(IncompleteDirectiveError, 'unknown variable named "%s"' % matches['value_var'])

      referred_var = variables[matches['value_var']]

      if isinstance(referred_var, integer_types):
        var.value = referred_var
      else:
        var.refers_to = referred_var

      return

    v_value = matches.get('value_label')
    if v_value:
      var.refers_to = v_value
      return

    raise buff.get_error(IncompleteDirectiveError, '.byte directive without a meaningful value')

  def __parse_ascii(var, matches):
    if not var.lineno:
      var.filename = buff.filename
      var.lineno = lineno

    matches = matches.groupdict()

    v_value = matches.get('value')
    if not v_value:
      raise buff.get_error(IncompleteDirectiveError, '.ascii directive without a string')

    DEBUG('Pre-decode: (%s) %s', type(v_value), ', '.join([str(ord(c)) for c in v_value]))
    s = decode_string(v_value)
    DEBUG('Pre-decode: (%s) %s', type(s), ', '.join([str(ord(c)) for c in s]))

    var.value = decode_string(v_value)

  def __parse_string(var, matches):
    if not var.lineno:
      var.filename = buff.filename
      var.lineno = lineno

    matches = matches.groupdict()

    v_value = matches.get('value')
    if not v_value:
      raise buff.get_error(IncompleteDirectiveError, '.string directive without a string')

    DEBUG('Pre-decode: (%s) %s', type(v_value), ', '.join([str(ord(c)) for c in v_value]))
    s = decode_string(v_value)
    DEBUG('Pre-decode: (%s) %s', type(s), ', '.join([str(ord(c)) for c in s]))

    var.value = decode_string(v_value)

  def __parse_space(var, matches):
    if not var.lineno:
      var.filename = buff.filename
      var.lineno = lineno

    matches = matches.groupdict()

    if 'size' not in matches:
      raise buff.get_error(IncompleteDirectiveError, '.size directive without a size')

    var.size = int(matches['size'])

  def __handle_symbol_variable(v_name, v_type):
    if v_type == 'char':
      var = CharSlot()

    elif v_type == 'byte':
      var = ByteSlot()

    elif v_type == 'int':
      var = IntSlot()

    elif v_type == 'ascii':
      var = AsciiSlot()

    elif v_type == 'string':
      var = StringSlot()

    elif v_type == 'space':
      var = SpaceSlot()

    var.name = Label(v_name, curr_section, buff.filename, buff.lineno)
    var.filename = buff.filename
    var.lineno = buff.lineno

    while buff.has_lines():
      line = buff.get_line()

      if line is None:
        var.close()
        data_section.content.append(var)
        return

      matches = RE_COMMENT.match(line)
      if matches:
        continue

      msg_prefix = 'pass #1: {}:{}: '.format(os.path.split(buff.filename)[1], buff.lineno)

      line = __apply_defs(line)

      if not current_macro and __apply_macros(line):
        DEBUG(msg_prefix + 'macro replaced, get fresh line')
        continue

      matches = RE_TYPE.match(line)
      if matches:
        buff.put_line(line)
        break

      matches = RE_SIZE.match(line)
      if matches:
        matches = matches.groupdict()

        if 'size' not in matches:
          raise buff.get_error(IncompleteDirectiveError, '.size directive without a size')

        var.size = UInt16(int(matches['size']))
        continue

      matches = RE_INT.match(line)
      if matches:
        __parse_int(var, matches)
        continue

      matches = RE_ASCII.match(line)
      if matches:
        __parse_ascii(var, matches)
        continue

      matches = RE_STRING.match(line)
      if matches:
        __parse_string(var, matches)
        continue

      matches = RE_SPACE.match(line)
      if matches:
        __parse_space(var, matches)
        continue

      matches = RE_BYTE.match(line)
      if matches:
        __parse_byte(var, matches)
        continue

      buff.put_line(line)
      break

    var.close()
    data_section.content.append(var)

  labels = []
  variables = {}

  instruction_set = cpu.instructions.DuckyInstructionSet

  defs = collections.OrderedDict()

  macros = collections.OrderedDict()
  current_macro = None

  DEBUG('Pass #1: text section is .text')
  DEBUG('Pass #1: data section is .data')

  text_section = sections_pass1['.text']
  data_section = sections_pass1['.data']
  curr_section = text_section

  global_symbols = []

  ifs = []

  def __fast_forward():
    DEBUG(msg_prefix + 'fast forwarding')

    depth = 1

    while buff.has_lines():
      line = buff.get_line()

      if line is None:
        return

      if not line.strip():
        continue

      matches = RE_IFDEF.match(line)
      if matches:
        depth += 1
        continue

      matches = RE_IFNDEF.match(line)
      if matches:
        depth += 1
        continue

      matches = RE_ENDIF.match(line)
      if matches:
        depth -= 1
        if depth == 0:
          buff.put_line(line)
          return

      matches = RE_ELSE.match(line)
      if matches:
        depth -= 1
        if depth == 0:
          buff.put_line(line)
          return

  while buff.has_lines():
    line = buff.get_line()

    if line is None:
      break

    if not line.strip():
      continue

    msg_prefix = 'pass #1: {}:{}: '.format(os.path.split(buff.filename)[1], buff.lineno)

    line = __apply_defs(line)

    if not current_macro and __apply_macros(line):
      DEBUG(msg_prefix + 'macro replaced, get fresh line')
      continue

    matches = RE_COMMENT.match(line)
    if matches:
      continue

    msg_prefix = 'pass #1: {}:{}: '.format(os.path.split(buff.filename)[1], buff.lineno)

    matches = RE_IFDEF.match(line)
    if matches:
      var = matches.groupdict()['var']

      DEBUG(msg_prefix + 'ifdef %s', var)

      ifs.append((True, var))

      if var in defines:
        DEBUG(msg_prefix + 'defined, continue processing')
        continue

      __fast_forward()
      continue

    matches = RE_IFNDEF.match(line)
    if matches:
      var = matches.groupdict()['var']

      DEBUG(msg_prefix + 'ifndef %s', var)

      ifs.append((False, var))

      if var not in defines:
        DEBUG(msg_prefix + 'not defined, continue processing')
        continue

      __fast_forward()
      continue

    matches = RE_ENDIF.match(line)
    if matches:
      DEBUG(msg_prefix + 'removing the last conditional from stack: %s', ifs[-1])

      ifs.pop()
      continue

    matches = RE_ELSE.match(line)
    if matches:
      defined, var = ifs.pop()

      DEBUG(msg_prefix + 'previous block was "%s %s"', 'ifdef' if defined is True else 'ifndef', var)

      ifs.append((not defined, var))

      if defined and var in defines:
        __fast_forward()
        continue

      DEBUG(msg_prefix + 'continue processing')
      continue

    matches = RE_INCLUDE.match(line)
    if matches:
      matches = matches.groupdict()

      if 'file' not in matches:
        raise buff.get_error(IncompleteDirectiveError, '.include directive without path')

      DEBUG(msg_prefix + 'include: file=%s', matches['file'])

      replace = None

      for d in includes:
        filename = os.path.join(d, matches['file'])
        DEBUG(msg_prefix + '  checking file %s', filename)

        try:
          with open(filename, 'r') as f_in:
            replace = f_in.read()

        except IOError:
          DEBUG('    failed to read')

        else:
          DEBUG('    read as replacement')
          break

      if replace is None:
        raise buff.get_error(UnknownFileError, matches['file'])

      buff.put_buffer(replace, filename = filename)

      continue

    matches = RE_VAR_DEF.match(line)
    if matches:
      matches = matches.groupdict()

      v_name = matches.get('var_name')
      v_body = matches.get('var_body')

      if not v_name or not v_body:
        raise buff.get_error(IncompleteDirectiveError, 'bad variable definition')

      DEBUG(msg_prefix + 'variable defined: name=%s, value=%s', v_name, v_body)

      defs[re.compile(r'\${}'.format(v_name))] = v_body.strip()

      continue

    matches = RE_MACRO_DEF.match(line)
    if matches:
      matches = matches.groupdict()

      m_name = matches.get('macro_name')
      m_params = matches.get('macro_params')

      if not m_name:
        raise buff.get_error(IncompleteDirectiveError, 'bad macro definition')

      DEBUG(msg_prefix + 'macro defined: name=%s', m_name)

      if current_macro:
        raise buff.get_error(AssemblerError, 'overlapping macro definitions')

      current_macro = {
        'name':    m_name,
        'pattern': None,
        'params':  [p.strip() for p in m_params.strip().split(',')] if m_params else [],
        'body':    []
      }

      if current_macro['params'] and len(current_macro['params'][0]):
        arg_pattern = r'(?P<arg%i>(?:".*?")|(?:.*?))'
        arg_patterns = ',\s*'.join([arg_pattern % i for i in range(0, len(current_macro['params']))])

        current_macro['pattern'] = re.compile(r'^\s*\${}\s+{}\s*(?:[;/#].*)?$'.format(m_name, arg_patterns), re.MULTILINE)

      else:
        current_macro['pattern'] = re.compile(r'\s*\${}'.format(m_name))

      continue

    matches = RE_MACRO_END.match(line)
    if matches:
      if not current_macro:
        raise buff.get_error(AssemblerError, 'closing non-existing macro')

      macros[current_macro['pattern']] = current_macro
      DEBUG(msg_prefix + 'macro definition closed: name=%s', current_macro['name'])
      current_macro = None
      continue

    if current_macro:
      current_macro['body'].append(line)
      continue

    matches = RE_SECTION.match(line)
    if matches:
      matches = matches.groupdict()

      if 'name' not in matches:
        raise buff.get_error(IncompleteDirectiveError, '.section directive without section name')

      s_name = matches['name']

      if s_name not in sections_pass1:
        section_flags = SectionFlags.from_string(matches.get('flags') or '')
        section_type = SectionTypes.TEXT if section_flags.executable == 1 else SectionTypes.DATA

        section = sections_pass1[s_name] = Section(s_name, section_type, section_flags)
        DEBUG(msg_prefix + 'section %s created', s_name)

      curr_section = sections_pass1[s_name]

      if curr_section.type == SectionTypes.TEXT:
        text_section = curr_section
        DEBUG(msg_prefix + 'text section changed to %s', s_name)
      else:
        data_section = curr_section
        DEBUG(msg_prefix + 'data section changed to %s', s_name)

      continue

    matches = RE_DATA.match(line)
    if matches:
      matches = matches.groupdict()

      curr_section = data_section = sections_pass1[matches['name'] if 'name' in matches and matches['name'] else '.data']
      DEBUG(msg_prefix + 'data section is %s', data_section.name)
      continue

    matches = RE_TEXT.match(line)
    if matches:
      matches = matches.groupdict()

      curr_section = text_section = sections_pass1[matches['name'] if 'name' in matches and matches['name'] else '.text']
      DEBUG(msg_prefix + 'text section is %s', text_section.name)
      continue

    matches = RE_TYPE.match(line)
    if matches:
      matches = matches.groupdict()

      if 'type' not in matches:
        raise buff.get_error(IncompleteDirectiveError, '.type directive without a type')

      if 'name' not in matches:
        raise buff.get_error(IncompleteDirectiveError, '.type directive without a name')

      __handle_symbol_variable(matches['name'], matches['type'])

      continue

    matches = RE_BYTE.match(line)
    if matches:
      var = ByteSlot()
      __parse_byte(var, matches)

      if len(labels) > 1:
        raise buff.get_error(AssemblerError, 'Too many data labels: {}'.format(labels))

      var.name = labels[0] if labels else None
      var.close()

      DEBUG(msg_prefix + 'record byte value: name=%s, value=%s', var.name, var.value)
      data_section.content.append(var)

      labels = []
      continue

    matches = RE_INT.match(line)
    if matches:
      var = IntSlot()
      __parse_int(var, matches)

      if len(labels) > 1:
        raise buff.get_error(AssemblerError, 'Too many data labels: {}'.format(labels))

      var.name = labels[0] if labels else None
      var.close()

      DEBUG(msg_prefix + 'record int value: name=%s, value=%s, refers_to=%s', var.name, var.value, var.refers_to)
      data_section.content.append(var)

      labels = []
      continue

    matches = RE_ASCII.match(line)
    if matches:
      var = AsciiSlot()
      __parse_ascii(var, matches)

      if len(labels) > 1:
        raise buff.get_error(AssemblerError, 'Too many data labels: {}'.format(labels))

      var.name = labels[0] if labels else None
      var.close()

      DEBUG(msg_prefix + 'record ascii value: name=%s, value=%s', var.name, var.value)
      data_section.content.append(var)

      labels = []
      continue

    matches = RE_STRING.match(line)
    if matches:
      var = StringSlot()
      __parse_string(var, matches)

      if len(labels) > 1:
        raise buff.get_error(AssemblerError, 'Too many data labels: {}'.format(labels))

      var.name = labels[0] if labels else None
      var.close()

      DEBUG(msg_prefix + 'record string value: name=%s, value=%s', var.name, var.value)
      data_section.content.append(var)

      labels = []
      continue

    matches = RE_SPACE.match(line)
    if matches:
      var = SpaceSlot()
      __parse_space(var, matches)

      if len(labels) > 1:
        raise buff.get_error(AssemblerError, 'Too many data labels: {}'.format(labels))

      var.name = labels[0] if labels else None
      var.close()

      DEBUG(msg_prefix + 'record space: name=%s, value=%s', var.name, var.size)
      data_section.content.append(var)

      labels = []
      continue

    matches = RE_GLOBAL.match(line)
    if matches:
      matches = matches.groupdict()

      if 'name' not in matches:
        raise buff.get_error(IncompleteDirectiveError, '.global directive without variable')

      name = matches['name']

      global_symbols.append(name)
      continue

    matches = RE_SET.match(line)
    if matches:
      matches = matches.groupdict()

      if 'name' not in matches:
        raise buff.get_error(IncompleteDirectiveError, '.set directive without variable')

      name = matches['name']

      if matches.get('current'):
        value = (curr_section.name, UInt16(curr_section.ptr.u16))

      elif matches.get('value_dec'):
        value = int(matches['value_dec'])

      elif matches.get('value_hex'):
        value = int(matches['value_hex'], base = 16)

      elif matches.get('value_label'):
        value = matches['value_label']

      else:
        raise buff.get_error(IncompleteDirectiveError, '.set directive with unknown value')

      DEBUG(msg_prefix + 'set variable: name=%s, value=%s', name, value)
      variables[name] = value

      continue

    if line.endswith(':'):
      label = Label(line.strip()[:-1], curr_section, buff.filename, buff.lineno)
      labels.append(label)

      DEBUG(msg_prefix + 'record label: name=%s', label.name)
      continue

    DEBUG(msg_prefix + 'line: %s', line)

    # label, instruction, 2nd pass flags
    emited_inst = None

    # Find instruction descriptor
    line = line.strip()

    for desc in instruction_set.instructions:
      DEBUG(msg_prefix + 'pattern: %s', desc.pattern.pattern)
      if not desc.pattern.match(line):
        continue
      break

    else:
      raise buff.get_error(AssemblerError, 'Unknown pattern: line="{}"'.format(line))

    emited_inst = desc.emit_instruction(logger, line)
    emited_inst.desc = desc

    if labels:
      text_section.content.append((labels, emited_inst))

    else:
      text_section.content.append((None, emited_inst))

    labels = []

    DEBUG(msg_prefix + 'emitted instruction: %s', emited_inst.desc.instruction_set.disassemble_instruction(emited_inst))

    if isinstance(desc, cpu.instructions.Inst_SIS):
      DEBUG(msg_prefix + 'switching istruction set: inst_set=%s', emited_inst.immediate)

      instruction_set = cpu.instructions.get_instruction_set(emited_inst.immediate)

  for s_name, section in iteritems(sections_pass1):
    DEBUG('pass #1: section %s', s_name)

    if section.type == SectionTypes.TEXT:
      for labeled, inst in section.content:
        DEBUG('pass #1: inst=%s, labeled=%s', inst, labeled)

    else:
      for var in section.content:
        DEBUG('pass #1: %s', var)

  DEBUG('Pass #2')

  sections_pass2 = collections.OrderedDict()
  references = {}
  base_ptr = UInt16(base_address.u16)

  for s_name, p1_section in iteritems(sections_pass1):
    section = sections_pass2[s_name] = Section(s_name, p1_section.type, p1_section.flags)

  symtab = sections_pass2['.symtab']
  reloctab = sections_pass2['.reloc']

  for s_name, section in iteritems(sections_pass2):
    p1_section = sections_pass1[s_name]

    section.base = UInt16(base_ptr.u16)
    section.ptr  = UInt16(base_ptr.u16)

    DEBUG('pass #2: section %s - base=%s', section.name, ADDR_FMT(section.base.u16))

    if section.type == SectionTypes.SYMBOLS or section.type == SectionTypes.RELOC:
      continue

    if section.type == SectionTypes.DATA:
      for var in p1_section.content:
        ptr_prefix = 'pass #2: ' + ADDR_FMT(section.ptr.u16) + ': '

        DEBUG(ptr_prefix + str(var))

        if var.name:
          var.section = section
          var.section_ptr = UInt16(section.ptr.u16)
          references['&' + var.name.name] = var

          symtab.content.append(var)

        if var.refers_to:
          refers_to = var.refers_to

          if isinstance(refers_to, tuple):
            reloc = RelocSlot(refers_to[0], patch_section = section, patch_address = section.ptr.u16, patch_offset = 0, patch_size = 16)
            DEBUG(ptr_prefix + 'reloc slot created: %s', reloc)

          else:
            reloc = RelocSlot(refers_to[1:], patch_section = section, patch_address = section.ptr.u16, patch_offset = 0, patch_size = 16)
            DEBUG(ptr_prefix + 'reloc slot created: %s', reloc)

          reloctab.content.append(reloc)
          var.refers_to = None

        if type(var) == IntSlot:
          if var.value:
            section.content.append(UInt8(var.value.u16 & 0x00FF))
            section.content.append(UInt8((var.value.u16 & 0xFF00) >> 8))
            DEBUG(ptr_prefix + 'value stored')

          else:
            section.content.append(var)
            DEBUG(ptr_prefix + 'value missing - reserve space, fix in next pass')

          section.ptr.u16 += 2

        elif type(var) == ByteSlot:
          section.content.append(UInt8(var.value.u8))
          section.ptr.u16 += var.size.u16
          DEBUG(ptr_prefix + 'value stored')

        elif type(var) == AsciiSlot or type(var) == StringSlot:
          for i in range(0, var.size.u16):
            section.content.append(var.value[i])
            section.ptr.u16 += 1

          if var.size.u16 % 2 != 0:
            section.content.append(UInt8(0))
            section.ptr.u16 += 1

          DEBUG(ptr_prefix + 'value stored')

        elif type(var) == SpaceSlot:
          section.content.append(var)
          section.ptr.u16 += var.size.u16
          DEBUG(ptr_prefix + 'value stored')

    if section.type == SectionTypes.TEXT:
      for labeled, inst in p1_section.content:
        ptr_prefix = 'pass #2: ' + ADDR_FMT(section.ptr.u16) + ': '

        inst.address = UInt16(section.ptr.u16)

        if labeled:
          for label in labeled:
            var = FunctionSlot()
            var.name = label
            var.section = section
            var.section_ptr = UInt16(section.ptr.u16)

            var.filename = label.filename
            var.lineno = label.lineno

            var.close()

            symtab.content.append(var)

            references['&' + label.name] = var
            DEBUG(ptr_prefix + 'label entry "%s" created', label)

        if inst.desc.operands and ('i' in inst.desc.operands or 'j' in inst.desc.operands) and hasattr(inst, 'refers_to') and inst.refers_to:
          reloc = RelocSlot(inst.refers_to[1:], flags = RelocFlags(relative = inst.desc.relative_address), patch_section = section, patch_address = section.ptr.u16)
          inst.desc.fill_reloc_slot(logger, inst, reloc)
          sections_pass2['.reloc'].content.append(reloc)

          if inst.refers_to in references:
            reloc.patch_section = references[inst.refers_to].section

          DEBUG(ptr_prefix + 'reloc slot created: %s', reloc)
          inst.refers_to = None

        section.content.append(inst)
        DEBUG(ptr_prefix + inst.desc.instruction_set.disassemble_instruction(inst))
        section.ptr.u16 += 4

    base_ptr.u16 = align_to_next_mmap(section.ptr.u16) if mmapable_sections else align_to_next_page(section.ptr.u16)

  DEBUG('Pass #3')

  sections_pass3 = {}

  for s_name, p2_section in iteritems(sections_pass2):
    section = Section(s_name, p2_section.type, p2_section.flags)
    sections_pass3[s_name] = section
    section.base = UInt16(p2_section.base.u16)
    section.ptr  = UInt16(section.base.u16)

  symtab = sections_pass3['.symtab']
  reloctab = sections_pass3['.reloc']

  for s_name, section in iteritems(sections_pass3):
    DEBUG('pass #3: section %s', section.name)

    p2_section = sections_pass2[s_name]

    if section.type == SectionTypes.SYMBOLS:
      symtab = section

    elif section.type == SectionTypes.RELOC:
      reloctab = section

    for item in p2_section.content:
      ptr_prefix = 'pass #3: ' + ADDR_FMT(section.ptr.u16) + ': '

      if section.type == SectionTypes.SYMBOLS:
        if (type(item.name) is Label and item.name.name in global_symbols) or item.name in global_symbols:
          item.flags.globally_visible = 1

      elif type(item) == IntSlot:
        if item.refers_to:
          refers_to = item.refers_to

          if isinstance(refers_to, tuple):
            referred_section = sections_pass2[item.refers_to[0]]

            reloc = RelocSlot(refers_to[0], patch_section = referred_section, patch_address = section.ptr.u16, patch_offset = 0, patch_size = 16)
            DEBUG(ptr_prefix + 'reloc slot created: %s', reloc)

          else:
            reloc = RelocSlot(refers_to[1:], patch_section = references[item.refers_to].section, patch_address = section.ptr.u16, patch_offset = 0, patch_size = 16)
            DEBUG(ptr_prefix + 'reloc slot created: %s', reloc)

          reloctab.content.append(reloc)

          item.value = 0x7979
          item.refers_to = None

        item.close()

        item = [UInt8(item.value.u16 & 0x00FF), UInt8((item.value.u16 & 0xFF00) >> 8)]

      elif hasattr(item, 'refers_to') and item.refers_to:
        DEBUG(ptr_prefix + 'fix reference: %s', item)

        reloc = RelocSlot(item.refers_to[1:], flags = RelocFlags(relative = item.desc.relative_address), patch_section = section, patch_address = section.ptr.u16)
        item.desc.fill_reloc_slot(logger, item, reloc)
        DEBUG(ptr_prefix + 'reloc slot created: %s', reloc)
        reloctab.content.append(reloc)

      DEBUG(ptr_prefix + str(item))

      if not isinstance(item, list):
        item = [item]

      for i in item:
        section.content.append(i)
        section.ptr.u16 += sizeof(i)

    DEBUG('pass #3: section %s finished: %s', section.name, section)

  DEBUG('Bytecode sections:')
  for s_name, section in iteritems(sections_pass3):
    DEBUG(str(section))

  DEBUG('Bytecode translation completed')

  return sections_pass3



import enum

from ...interfaces import ISnapshotable
from ..import CPUException
from . import Coprocessor
from ...mm import SEGMENT_SHIFT
from ...errors import AccessViolationError

from ctypes import c_ushort as u16

class ReadOnlyRegisterError(CPUException):
  def __init__(self, r, *args, **kwargs):
    super(ReadOnlyRegisterError, self).__init__('Register cr{:i} is read-only.'.format(r), *args, **kwargs)

class WriteOnlyRegisterError(CPUException):
  def __init__(self, r, *args, **kwargs):
    super(WriteOnlyRegisterError, self).__init__('Register cr{:i} is write-only.'.format(r), *args, **kwargs)

class ControlRegisters(enum.IntEnum):
  CR0 = 0  # System ID (reserved)
  CR1 = 1  # Interrupt Vector Table address
  CR2 = 2  # Interrupt Vector Table segment

class MathCoprocessor(ISnapshotable, Coprocessor):
  def read_cr0(self):
    return u16(0)

  def write_cr0(self, value):
    raise ReadOnlyRegisterError('cr0')

  def read_cr1(self):
    return u16(self.core.ivt_address & 0xFFFF)

  def write_cr1(self, address):
    self.core.ivt_address = (self.core.ivt_address & 0xFF0000) | address

  def read_cr2(self):
    return u16((self.core.ivt_address >> SEGMENT_SHIFT) & 0xFF)

  def write_cr2(self, segment):
    self.core.ivt_address = (self.core.ivt_address & 0x00FFFF) | ((segment & 0xFF) << SEGMENT_SHIFT)

  def read(self, r):
    if not self.core.privileged:
      raise AccessViolationError('It is not allowed to read control registers in non-privileged mode')

    handler = 'read_cr%i' % r

    if not hasattr(self, handler):
      raise WriteOnlyRegisterError(r)

    return getattr(self, handler)()

  def write(self, r, value):
    if not self.core.privileged:
      raise AccessViolationError('It is not allowed to modify control registers in non-privileged mode')

    handler = 'write_cr%i' % r

    if not hasattr(self, handler):
      raise ReadOnlyRegisterError(r)

    return getattr(self, handler)(value)



"""
Keyboard controller - provides events for pressed and released keys.
"""

import enum
import io

from . import IRQProvider, IOProvider, Device, IRQList
from ..errors import InvalidResourceError
from ..mm import UINT16_FMT

DEFAULT_PORT_RANGE = 0x100


class ControlMessages(enum.IntEnum):
  HALT = 1025

  CONTROL_MESSAGE_FIRST = 1024


def escape_char(c):
  return chr(c).replace(chr(10), '\\n').replace(chr(13), '\\r')


class KeyboardController(IRQProvider, IOProvider, Device):
  def __init__(self, machine, name, streams = None, port = None, irq = None, *args, **kwargs):
    super(KeyboardController, self).__init__(machine, 'input', name, *args, **kwargs)

    self.streams = streams[:] if streams else []
    self.port = port or DEFAULT_PORT_RANGE
    self.ports = [port]
    self.irq = irq or IRQList.KEYBOARD

    self.input = None

    self.queue = []

  @staticmethod
  def create_from_config(machine, config, section):
    return KeyboardController(machine,
                              section,
                              streams = None,
                              port = config.getint(section, 'port', DEFAULT_PORT_RANGE),
                              irq = config.getint(section, 'irq', IRQList.KEYBOARD))

  def __repr__(self):
    return 'basic keyboard controller on [%s] as %s' % (', '.join([UINT16_FMT(port) for port in self.ports]), self.name)

  def enqueue_input(self, stream):
    self.machine.DEBUG('KeyboardController.enqueue_input: stream=%s', stream)

    if not stream.has_fd():
      raise InvalidResourceError('Keyboard controller requires input stream with fd support')

    self.streams.append(stream)

  def close_input(self):
    self.machine.DEBUG('KeyboardController.close_input: input=%s', self.input)

    if self.input is None:
      return

    self.machine.reactor.remove_fd(self.input.fd)

    self.input.close()
    self.input = None

  def open_input(self):
    self.machine.DEBUG('KeyboardController.open_input')

    self.close_input()

    if not self.streams:
      self.machine.DEBUG('no additional input streams')
      if not self.queue or self.queue[-1] != ControlMessages.HALT:
        self.machine.DEBUG('signal halt')
        self.queue.append(ControlMessages.HALT)
      return

    self.input = self.streams.pop(0)
    self.machine.DEBUG('KeyboardController.input=%s', self.input)

    self.machine.reactor.add_fd(self.input.fd, on_read = self.handle_raw_input, on_error = self.handle_input_error)

  def boot(self):
    self.machine.DEBUG('KeyboardController.boot')

    for port in self.ports:
      self.machine.register_port(port, self)

    self.open_input()

    self.machine.INFO('hid: %s', self)

  def halt(self):
    self.machine.DEBUG('KeyboardController.halt')

    for port in self.ports:
      self.machine.unregister_port(port)

    self.close_input()

  def handle_input_error(self):
    self.machine.DEBUG('KeyboardController.handle_input_error')

    self.open_input()

  def handle_raw_input(self):
    self.machine.DEBUG('KeyboardController.handle_raw_input')

    assert self.input is not False

    buff = self.input.read(size = io.DEFAULT_BUFFER_SIZE)
    self.machine.DEBUG('KeyboardController.handle_raw_input: buff=%s (%s)', buff, type(buff))

    if buff is None:
      self.machine.DEBUG('KeyboardController.handle_raw_input: nothing to do, no input')
      return

    if not buff:
      # EOF
      self.open_input()
      return

    self.machine.DEBUG('KeyboardController.handle_raw_input: adding %i chars', len(buff))

    self.queue += buff

    self.machine.DEBUG('KeyboardController.handle_raw_input: queue now has %i bytes', len(self.queue))

    self.machine.trigger_irq(self)

  def __read_char(self):
    self.machine.DEBUG('KeyboardController.__read_char')

    while True:
      try:
        b = self.queue.pop(0)

      except IndexError:
        self.machine.DEBUG('KeyboardController.__read_char: no available chars in queue')
        return None

      self.machine.DEBUG('KeyboardController.__read_char: queue now has %i bytes', len(self.queue))

      if b == ControlMessages.HALT:
        self.machine.DEBUG('KeyboardController.__read_char: planned halt, execute')
        self.machine.halt()

        return None

      break

    self.machine.DEBUG('KeyboardController.__read_char: c=%s ()', b)

    return b

  def read_u8(self, port):
    self.machine.DEBUG('KeyboardController.read_u8: port=%s', UINT16_FMT(port))

    if port not in self.ports:
      raise InvalidResourceError('Unhandled port: %s', UINT16_FMT(port))

    b = self.__read_char()
    if not b:
      self.machine.DEBUG('KeyboardController.read_u8: empty input, signal it downstream')
      return 0xFF

    self.machine.DEBUG('KeyboardController.read_u8: input byte is %i', b)

    return b
